# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module to generate the circuits necessary to implement linear combinations of unitaries
Refs:
    [1] Dominic W. Berry, Andrew M. Childs, Richard Cleve, Robin Kothari, Rolando D. Somma, "Simulating
    Hamiltonian dynamics with a truncated Taylor series" arXiv: 1412.4687 Phys. Rev. Lett. 114, 090502 (2015)
"""

import math
from typing import Union, Tuple, List

import numpy as np

from tangelo.linq import Gate, Circuit
from tangelo.linq.helpers.circuits.statevector import StateVector
from tangelo.toolboxes.operators.operators import QubitOperator, count_qubits


def get_truncated_taylor_series(qu_op: QubitOperator, kmax: int, t: float, control: Union[int, List[int]] = None) -> Circuit:
    r"""Generate Circuit to implement the truncated Taylor series algorithm as implemented in arXiv:1412.4687
    Args:
        qu_op (QubitOperator): The qubit operator to apply the truncated Taylor series exponential
        kmax (int): The maximum order of the Taylor series \exp{-1j*t*qu_op} = \sum_{k}^{kmax}(-1j*t)**k/k! qu_op**k
        t (float): The total time to evolve
        control (int or list[int]): The control qubit(s)
    Returns:
        Circuit: the circuit that implements the time-evolution of qu_op for time t with Taylor series kmax
    """

    if kmax < 1:
        raise ValueError("Taylor series can only be applied for kmax > 0")

    qu_op_size = count_qubits(qu_op)

    kprep, unitaries, rsteps = Uprepkl(qu_op, kmax, t)

    kprep_qubits = list(range(qu_op_size, qu_op_size + kprep.width))
    kprep.reindex_qubits(kprep_qubits)

    kselect = USelectkl(unitaries, qu_op_size, kmax, control)
    flip_op = sign_flip(kprep_qubits, control)

    lcu_circuit = kprep + kselect + kprep.inverse()

    amplified_lcu_circuit = lcu_circuit + flip_op + lcu_circuit.inverse() + flip_op + lcu_circuit

    # Added gates below because current implementation applies -1j*exp(-1j*H*t) and global phase
    # matters for controlled operations
    # TODO: Find a way to incorporate this phase into the time propagation natively.
    if control is not None:
        gates = [Gate("CRZ", q, control=control, parameter=np.pi/2) for q in range(qu_op_size)]
        gates += [Gate("CPHASE", q, control=control, parameter=-np.pi/2) for q in range(qu_op_size)]
        amplified_lcu_circuit += Circuit(gates)

    return amplified_lcu_circuit * rsteps


def Uprepkl(qu_op: QubitOperator, kmax: int, t: float) -> Tuple[Circuit, List[QubitOperator], int]:
    """Generate Uprep circuit using qubit encoding defined in arXiv:1412.4687
    Args:
        qu_op (QubitOperator) :: The qubit operator to obtain the Uprep circuit for
        kmax (int): the order of the truncated Taylor series
        t (float): The evolution time
    Returns:
        Circuit: The Uprep circuit for the truncated Taylor series
        list: the individual QubitOperator unitaries with only the prefactor remaining. i.e.  all coefficients are 1, -1, 1j or -1j
        int: The number of repeated applications of the circuit that need to be applied
    """

    # Incorporate sign of time into qubit operator.
    qu_op_new = np.sign(t) * qu_op
    t_new = abs(t)

    # remove the coefficient value and generate the list of QubitOperator with only its phase 1j, or -1j
    vector = list()
    unitaries = list()
    for term, coeff in qu_op_new.terms.items():
        if np.abs(coeff.imag) > 1.e-7:
            raise ValueError(f"Only real qubit operators are allowed but term {term} has coefficient {coeff}")
        if coeff.real > 0:
            unitaries.append(QubitOperator(term, -1j))
        else:
            unitaries.append(QubitOperator(term, 1j))
        vector.append(np.abs(coeff.real))

    num_terms = len(vector)
    vector = np.array(vector)
    vsum = sum(vector)

    # Calculate 1-norm of coefficients in qubit operator and obtain the maximum time-step allowed
    # These values are obtained by finding the relevant root of the kth order Taylor series polynomial approximation
    # of 2 = exp(x) (i.e. the roots of 2 = \sum_{k=0}^N x^k/k!). The limit for large k is log(2)
    poly_roots = {1: 1., 2: 0.73205081, 3: 0.69888549, 4: 0.69390315, 5: 0.69323260, 6: 0.69315552, 7: 0.69314790, 8: 0.69314724}
    max_time_step = poly_roots[kmax] / vsum if kmax < 9 else np.log(2)

    # Calculate the number of time steps required and calculate the actual 1-norm for each time-step
    time_steps = math.ceil(t_new / max_time_step)
    vsum_with_t = sum(coeff * t_new / time_steps for coeff in vector)
    expvsum = sum((vsum_with_t) ** k / math.factorial(k) for k in range(0, kmax+1))

    # Generate vector v_i = sqrt(alpha_i)/np.sqrt(1-norm) with zeros padded to create a vector of length 2**n
    vector = np.sqrt(vector) / np.sqrt(vsum)
    n_qubits = math.ceil(math.log2(num_terms))
    newvec = np.zeros(2 ** n_qubits)
    newvec[:num_terms] = vector

    # Calculate circuit that generates vector and apply controls for Taylor series
    s = StateVector(newvec, order="msq_first")
    qss = list()
    for k in range(kmax):
        qss.append(s.initializing_circuit())
        # Reindex to correct block of qubits for kth order of Taylor series
        qss[-1].reindex_qubits(list(range(kmax + k * n_qubits, kmax + (k+1) * n_qubits)))
        # Add control for the kth unary encoded qubits
        for gate in qss[-1]._gates:
            if gate.control is not None:
                gate.control += [k]
            else:
                gate.name = "C"+gate.name
                gate.control = [k]
    qstot = Circuit()
    for qs in qss:
        qstot += qs

    # Calculate coefficients for unary encoding k="0"+"0"*(kmax-k)*"1"*k where
    # first qubit encodes extra identity term needed to ensure 1-norm = 2 such that
    # oblivious amplitude amplification is applicable
    kvec = np.zeros(kmax + 2)
    for k in range(kmax+1):
        # for position "0" + "0"*(kmax-k) + "1"*k
        kvec[k] = (t_new / time_steps * vsum) ** k / math.factorial(k)
    kvec[0] += ((2 - expvsum) / 2)
    # for position "1" + "0" * kmax
    kvec[kmax+1] = ((2 - expvsum) / 2)

    kvec = np.sqrt(kvec) / np.sqrt(np.sum(kvec))

    kprep = get_unary_prep(kvec, kmax)

    # shift encoding of extra identity term to last qubit
    kprep.reindex_qubits(list(range(kmax)) + [kmax + kmax * n_qubits])

    return kprep + qstot, unitaries, time_steps


def get_unary_prep(kvec: np.ndarray, kmax: int) -> Circuit:
    """Generate the prep circuit for the unary+ancilla part of the Taylor series encoding. This implementation
    scales linearly with kmax whereas StateVector.initializing_circuit() scaled exponentially.

    Args:
        kvec (array): Array representing the coefficients needed to generate unary portion of encoding.
            Length of array is kmax+2. order "00...000", "00...001", "00...011", ..., "01...111", "1000..."
        kmax (int): The Taylor series order

    Returns:
        Circuit : The unary encoding prep circuit
    """

    # Generate ancilla value and apply "X" so control is on other portion
    # For "1" + kmax*"0"
    val = kvec[kmax + 1]
    gates = [Gate("RY", kmax, parameter=np.arcsin(val)*2), Gate("X", kmax)]

    # Keep track of remaining value in constant c
    c = np.cos(np.arcsin(val))
    for i in range(0, kmax):
        # Obtain new value to generate and rotate by np.arccos(val/c) for "0" + "0"*(kmax-i) + "1"*i
        val = kvec[i]
        control = [kmax] + [i-1] if i > 0 else [kmax]
        gates += [Gate("CRY", i, control=control, parameter=np.arccos(val/c)*2)]
        c *= np.sin(np.arccos(val/c))
    gates += [Gate("X", kmax)]

    return Circuit(gates)


def USelectkl(unitaries: List[QubitOperator], n_qubits_sv: int, kmax: int, control: Union[int, List[int]] = None) -> Circuit:
    r"""Generate the truncated Taylor series U_{Select} circuit for the list of QubitOperator as defined arXiv:1412.4687
    The returned Circuit will have qubits defined in registers |n_qubits_sv>|kmax-1>|n_qubits_u>^{kmax-1}|ancilla>
    n_qubits_sv is the number of qubits to define the state to propagate. |kmax-1> defines the unary encoding
    of the Taylor series order. kmax-1 copies of length log2(len(unitaries)) to define the binary encoding of the linear
    combination of unitaries. Finally, one ancilla qubit to ensure the success of oblivious amplitude amplification.
    Args:
        unitaries (list[QubitOperator]): The list of unitaries that defines the U_{Select} operation
        n_qubits_sv (int): The number of qubits in the statevector register
        kmax (int): The maximum Taylor series order
        control (int or list[int]): Control qubits for operation
    Returns:
        Circuit: The circuit that implements the truncated Taylor series U_{Select} operation
    """

    n_qubits_u = math.ceil(math.log2(len(unitaries)))
    if control is not None:
        control_list = control if isinstance(control, list) else [control]
    else:
        control_list = []

    gate_list = []
    for k in range(1, kmax+1):
        q_start = n_qubits_sv + kmax + (k-1) * n_qubits_u
        k_control_qubits = [n_qubits_sv + k - 1] + list(range(q_start, q_start + n_qubits_u)) + control_list

        for j, unitary in enumerate(unitaries):
            bs = bin(j).split('b')[-1]
            state_binstr = "0" * (n_qubits_u - len(bs)) + bs
            state_binstr = state_binstr[::-1]

            x_ladder = [Gate("X", q_start + q) for q, i in enumerate(state_binstr) if i == "0"]

            # Add phase to controlled-unitary
            gate_list += x_ladder
            for term, coeff in unitary.terms.items():
                phasez = np.arctan2(-coeff.imag, coeff.real)
                phasep = -phasez*2
                if abs(phasez) > 1.e-12:
                    gate_list.append(Gate("CRZ", target=0, control=k_control_qubits, parameter=phasez*2))
                if abs(phasep) > 1.e-12 or not np.isclose(abs(phasep), np.pi*2):
                    gate_list.append(Gate("CPHASE", target=0, control=k_control_qubits, parameter=phasep))
                gate_list += [Gate("C"+op, target=index, control=k_control_qubits) for index, op in term]
            gate_list += x_ladder

    # Add -I term to ensure total sum equals 2 for oblivious amplitude amplification
    k_control_qubits = [n_qubits_sv+ki for ki in range(kmax)] + [n_qubits_sv + kmax + kmax*n_qubits_u] + control_list
    x_ladder = [Gate("X", n_qubits_sv + q) for q in range(kmax)]
    gate_list += x_ladder
    gate_list.append(Gate("CRZ", target=0, control=k_control_qubits, parameter=2*np.pi))
    gate_list += x_ladder

    return Circuit(gate_list)


def sign_flip(qubit_list: List[int], control: Union[int, List[int]] = None) -> Circuit:
    """Generate Circuit corresponding to the sign flip of the |0>^n vector for the given qubit_list
    Args:
        qubit_list (list[int]): The list of n qubits for which the 2*|0>^n-I operation is generated
        control (int or list[int]): Control qubit or list of control qubits.
    Returns:
        Circuit: The circuit that generates the sign flip on |0>^n
    """
    if control is not None:
        fcontrol_list = control if isinstance(control, list) else [control]
    else:
        fcontrol_list = []
    gate_list = []

    x_ladder = [Gate("X", q) for q in qubit_list]
    fcontrol_list += qubit_list[:-1]

    gate_list += x_ladder
    gate_list.append(Gate('CZ', target=qubit_list[-1], control=fcontrol_list))
    gate_list += x_ladder
    return Circuit(gate_list)


def get_lcu_circuit(qu_op: QubitOperator, control: Union[int, List[int]] = None) -> Circuit:
    """Apply qu_op using the linear combination of unitaries algorithm
    1-norm of coefficients must be less than 2. The unitarity of qu_op is not checked by the algorithm
    Args:
        qu_op (QubitOperator): The qu_op to apply. Must be nearly unitary for algorithm to succeed with high probability.
        control (int or list[int]): Control qubit(s)
    Returns:
        Circuit: The circuit that implements the linear combination of unitaries for the qu_op
    """

    if control is not None:
        control_list = control if isinstance(control, list) else [control]
    else:
        control_list = []

    unitaries = list()
    vector = list()
    max_qu_op = count_qubits(qu_op)
    for term, coeff in qu_op.terms.items():
        acoeff = np.abs(coeff)
        vector += [acoeff]
        unitaries += [QubitOperator(term, coeff / acoeff)]
    vsum = sum(vector)

    # Check that 1-norm is less than 2 and add and subtract terms proportional to the identity to obtain 1-norm equal to 2
    if vsum > 2 + 1.e-10:
        raise ValueError("Can not apply qu_op as its 1-norm is greater than 2")
    vector += [max((2 - vsum)/2, 0), max((2 - vsum)/2, 0)]
    unitaries += [QubitOperator(tuple(), 1), QubitOperator(tuple(), -1)]

    # create U_{prep} from sqrt of coefficients
    num_terms = len(vector)
    vector = np.array(vector)
    vector = np.sqrt(vector)/np.sqrt(2)
    n_qubits = math.ceil(math.log2(num_terms))
    newvec = np.zeros(2**n_qubits)
    newvec[0:num_terms] = vector[:]
    s = StateVector(newvec, order="lsq_first")
    uprep = s.initializing_circuit()
    uprep_qubits = list(range(max_qu_op, max_qu_op + n_qubits))
    uprep.reindex_qubits(uprep_qubits)

    # Generate U_{Select}
    gate_list = list()
    control_qubits = uprep_qubits + control_list
    for j, unitary in enumerate(unitaries):
        bs = bin(j).split('b')[-1]
        state_binstr = "0" * (n_qubits - len(bs)) + bs
        x_ladder = [Gate("X", q + max_qu_op) for q, i in enumerate(state_binstr) if i == "0"]
        gate_list += x_ladder
        for term, coeff in unitary.terms.items():
            phasez = np.arctan2(-coeff.imag, coeff.real)
            phasep = -phasez*2

            if abs(phasez) > 1.e-12:
                gate_list.append(Gate("CRZ", target=0, control=control_qubits, parameter=phasez*2))
            if abs(phasep) > 1.e-12 or not np.isclose(abs(phasep), np.pi*2):
                gate_list.append(Gate("CPHASE", target=0, control=control_qubits, parameter=phasep))
            gate_list += [Gate("C"+op, target=index, control=control_qubits) for index, op in term]
        gate_list += x_ladder

    uselect = Circuit(gate_list)

    flip_op = sign_flip(uprep_qubits, control=control)
    w = uprep + uselect + uprep.inverse()

    amplified_lcu_circuit = w + flip_op + w.inverse() + flip_op + w

    # Added gates below because current implementation applies -1j*exp(-1j*H*t) and global phase
    # matters for controlled operations
    # TODO: Find a way to incorporate this phase into the time propagation natively.
    if control is not None:
        gates = [Gate("CRZ", q, control=control, parameter=np.pi/2) for q in range(max_qu_op)]
        gates += [Gate("CPHASE", q, control=control, parameter=-np.pi/2) for q in range(max_qu_op)]
        amplified_lcu_circuit += Circuit(gates)

    return amplified_lcu_circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import numpy as np

from tangelo.toolboxes.molecular_computation.molecule import Molecule, SecondQuantizedMolecule

sup = np.testing.suppress_warnings()
warnings.filterwarnings("ignore", message="Using default_file_mode other than 'r' is deprecated")
warnings.filterwarnings("ignore", message="`np")
warnings.filterwarnings("ignore", category=DeprecationWarning)
sup.filter(np.core)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines functions to get suggestions for freezing orbitals. Those
functions take a molecule and return an integer or a list of orbital indexes for
freezing orbitals. Depending on the function, a Molecule or a
SecondQuantizedMolecule object can be used.
"""


def get_frozen_core(molecule):
    """Function to compute the number of frozen orbitals. This function is only
    for the core (occupied orbitals).

    Args:
        molecule (SecondQuantizedMolecule): Molecule to be evaluated.

    Returns:
        int: First N molecular orbitals to freeze.
    """

    # Freezing core of each atom. "Row" refers to a periodic table row.
    # 1st row: None are frozen.
    # 2nd row: 1s is frozen.
    # 3rd row: 1s, 2s, 2px, 2py and 2pz are frozen.
    core_orbitals = {
        "H": 0, "He": 0,
        "Li": 1, "Be": 1, "B": 1, "C": 1, "N": 1, "O": 1, "F": 1, "Ne": 1,
        "Na": 5, "Mg": 5, "Al": 5, "Si": 5, "P": 5, "S": 5, "Cl": 5, "Ar": 5
    }

    # Counting how many of each element is in the molecule.
    elements = {i: molecule.elements.count(i) for i in molecule.elements}
    frozen_core = sum([v * core_orbitals.get(k, 0) for k, v in elements.items()])

    return frozen_core


def get_orbitals_excluding_homo_lumo(molecule, homo_minus_n=0, lumo_plus_n=0):
    """Function that returns a list of orbitals to freeze if the user wants to
    consider only a subset from HOMO(-homo_min_n) to LUMO(+lumo_plus_n)
    orbitals. Users should be aware of degeneracies, as this function does not
    take this property into account.

    Args:
        molecule (SecondQuantizedMolecule): Molecule to be evaluated.
        homo_minus_n (int): Starting point at HOMO - homo_minus_n.
        lumo_plus_n (int): Ending point at LUMO + lumo_plus_n.

    Returns:
        list of int: Frozen orbitals not detected in the active space.
    """

    # Getting the number of molecular orbitals. It also works with different
    # basis sets.
    n_molecular_orb = molecule.n_mos

    n_lumo = molecule.mo_occ.tolist().index(0.)
    n_homo = n_lumo - 1

    frozen_orbitals = [n for n in range(n_molecular_orb) if n not in range(n_homo-homo_minus_n, n_lumo+lumo_plus_n+1)]

    return frozen_orbitals



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions helping with quantum circuit format conversion between abstract
format and qdk/qsharp format.

In order to produce an equivalent circuit for the target backend, it is
necessary to account for:
- how the gate names differ between the source backend to the target backend.
- how the order and conventions for some of the inputs to the gate operations
    may also differ.
"""


def get_qdk_gates():
    """Map gate name of the abstract format to the equivalent gate name used in
    Q# operations API and supported gates:
    https://docs.microsoft.com/en-us/qsharp/api/qsharp/microsoft.quantum.intrinsic
    """

    GATE_QDK = dict()
    for name in {"H", "X", "Y", "Z", "S", "T", "CNOT"}:
        GATE_QDK[name] = name
    for name in {"RX", "RY", "RZ"}:
        GATE_QDK[name] = name[0] + name[1:].lower()
    GATE_QDK["PHASE"] = "R1"
    GATE_QDK["CPHASE"] = "R1"
    for name in {"CRX", "CRY", "CRZ"}:
        GATE_QDK[name] = name[1] + name[2:].lower()
    for name in {"CH", "CX", "CY", "CZ", "CS", "CT", "CSWAP"}:
        GATE_QDK[name] = name[1:]
    GATE_QDK["SWAP"] = "SWAP"
    GATE_QDK["MEASURE"] = "M"

    return GATE_QDK


def translate_qsharp(source_circuit, operation="MyQsharpOperation"):
    """Take in an abstract circuit, generate the corresponding Q# operation
    (state prep + measurement) string, in the appropriate Q# template. The Q#
    output can be written to file and will be compiled at runtime.

    Args:
        source_circuit: quantum circuit in the abstract format.
        operation (str), optional: name of the Q# operation.

    Returns:
        str: The Q# code (operation + template). This needs to be written into a
            .qs file, and compiled at runtime.
    """

    GATE_QDK = get_qdk_gates()

    # Prepare Q# operation header
    qsharp_string = ""
    qsharp_string += "@EntryPoint()\n"
    qsharp_string += f"operation {operation}() : Result[] {{\n"
#    qsharp_string += "body (...) {\n\n"
    qsharp_string += f"\tmutable c = new Result[{source_circuit.width}];\n"
    qsharp_string += f"\tusing (qreg = Qubit[{source_circuit.width}]) {{\n"

    # Generate Q# strings with the right syntax, order and values for the gate inputs
    body_str = ""
    for gate in source_circuit._gates:
        if gate.control is not None and gate.name != "CNOT":
            control_string = '['
            num_controls = len(gate.control)
            for i, c in enumerate(gate.control):
                control_string += f'qreg[{c}]]' if i == num_controls - 1 else f'qreg[{c}], '

        if gate.name in {"H", "X", "Y", "Z", "S", "T"}:
            body_str += f"\t\t{GATE_QDK[gate.name]}(qreg[{gate.target[0]}]);\n"
        elif gate.name in {"RX", "RY", "RZ", "PHASE"}:
            body_str += f"\t\t{GATE_QDK[gate.name]}({gate.parameter}, qreg[{gate.target[0]}]);\n"
        elif gate.name in {"CNOT"}:
            body_str += f"\t\t{GATE_QDK[gate.name]}(qreg[{gate.control[0]}], qreg[{gate.target[0]}]);\n"
        elif gate.name in {"CRX", "CRY", "CRZ", "CPHASE"}:
            body_str += f"\t\tControlled {GATE_QDK[gate.name]}({control_string}, ({gate.parameter}, qreg[{gate.target[0]}]));\n"
        elif gate.name in {"CH", "CX", "CY", "CZ", "CS", "CT"}:
            body_str += f"\t\tControlled {GATE_QDK[gate.name]}({control_string}, (qreg[{gate.target[0]}]));\n"
        elif gate.name in {"SWAP"}:
            body_str += f"\t\t{GATE_QDK[gate.name]}(qreg[{gate.target[0]}], qreg[{gate.target[1]}]);\n"
        elif gate.name in {"CSWAP"}:
            body_str += f"\t\tControlled {GATE_QDK[gate.name]}({control_string}, (qreg[{gate.target[0]}], qreg[{gate.target[1]}]));\n"
        elif gate.name in {"MEASURE"}:
            body_str += f"\t\tset c w/= {gate.target[0]} <- {GATE_QDK[gate.name]}(qreg[{gate.target[0]}]);\n"
        else:
            raise ValueError(f"Gate '{gate.name}' not supported on backend qdk")
    qsharp_string += body_str + "\n\t\treturn ForEach(MResetZ, qreg);\n"
    qsharp_string += "\t}\n"
#    qsharp_string += "}\n adjoint auto;\n"
    qsharp_string += "}\n"

    # Fills the template with the Q# operation. The written Q# file will be compiled by the qsharp module at runtime
    from .qdk_template import _qdk_template, _header
    template = _header + _qdk_template.format(operation_name=operation)

    # Return Q# code (template + operation) for simulation. Needs to be written to file and compiled at runtime
    return template + qsharp_string + "\n}\n"



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides access to different approaches aiming at reducing the
number of measurements required to compute the expectation value of a qubit
(Pauli) operator. They attempt to identify the minimal amount of measurement
bases / terms required to deduce the expectation values associated to the other
terms present in the Pauli operator.
"""

import warnings
from openfermion.measurements import group_into_tensor_product_basis_sets
from tangelo.linq import Simulator


def group_qwc(qb_ham, seed=None, n_repeat=1):
    """Wrapper around Openfermion functionality that takes as input a
    QubitOperator and yields a collection of measurement bases defining a
    partition of groups of sub-operators with terms that are diagonal in the
    same tensor product basis. Each sub-operator can be measured using the same
    qubit post-rotations in expectation estimation. This uses the idea of
    qubitwise commutativity (qwc), and the minimum clique cover algorithm.

    The resulting dictionary maps the measurement basis (key) to the list of
    qubit operators whose expectation value can be computed using the
    corresponding circuit. The size of this dictionary determines how many
    quantum circuits need to be executed in order to provide the expectation
    value of the input qubit operator.

    The minimum clique cover algorithm can be initialized with a random seed
    and can be repeated several times with different seeds in order to return
    the run that produces the lowest number of groups.

    Args:
        qb_ham (QubitOperator): the operator that will be split into
            sub-operators (tensor product basis sets).
        seed (int): default None. Random seed used to initialize the
            numpy.RandomState pseudo-RNG.
        n_repeat (int): Repeat with a different random seed, keep the outcome
            resulting in the lowest number of groups.

    Returns:
        dict: a dictionary where each key defines a tensor product basis, and
            each corresponding value is a QubitOperator with terms that are all
            diagonal in that basis.
    """

    res = group_into_tensor_product_basis_sets(qb_ham, seed)
    for i in range(n_repeat-1):
        res2 = group_into_tensor_product_basis_sets(qb_ham)
        if len(res2) < len(res):
            res = res2
    return res


def check_bases_commute_qwc(b1, b2):
    """ Check whether two bases commute qubitwise.

    Args:
        b1 (tuple of (int, str)): the first measurement basis
        b2 (tuple of (int, str)): the second measurement basis

    Returns:
        bool: whether or not the bases commute qubitwise
    """

    b1_dict, b2_dict = dict(b1), dict(b2)
    for i in set(b1_dict) & set(b2_dict):
        if b1_dict[i] != b2_dict[i]:
            return False
    return True


def map_measurements_qwc(qwc_group_map):
    """ Somewhat reverses a grouping dictionary, linking all measurement bases to the list of
    group representatives that commute with them qubitwise. Useful to find all the bases
    whose shots can be used in order to form data for other bases during a hardware experiment.

    Args:
        qwc_group_map (dict): maps a measurement basis to a qubit operator whose terms commute
            with said measurement basis qubit-wise.
    Returns:
        dict: maps each of the bases to all group representatives that commute with them qubitwise.
    """

    meas_map = dict()
    meas_bases, op_bases = list(), list()
    for k, v in qwc_group_map.items():
        meas_bases.append(k)
        op_bases.extend(v.terms.keys())

    for b1 in op_bases:
        for b2 in meas_bases:
            if b1 and check_bases_commute_qwc(b1, b2):
                meas_map[b1] = meas_map.get(b1, []) + [b2]

    return meas_map


def exp_value_from_measurement_bases(sub_ops, histograms):
    """Computes the expectation value of the sum of all suboperators
    corresponding to the different measurement bases in the input dictionary.
    This is how one would go about computing the expectation value of an
    operator while trying to send a limited amount of tasks to a quantum
    computer, thereby lowering the cost in terms of number of measurements
    needed.

    Args:
        sub_ops (dict): Maps measurement bases to suboperators whose exp. value
            can be computed from it.
        histograms (dict): Maps measurement bases to histograms corresponding to
            that quantum circuit.

    Returns:
        float or complex: Expectation value of the sum of all suboperators.
    """

    # Warning if dicts dont have exact set of keys
    if set(sub_ops) != set(histograms):
        warnings.warn("Measurement bases are not exactly the same in both dictionaries. Terms may be missing.")

    exp_value = 0.
    for basis, freqs in histograms.items():
        for term, coef in sub_ops[basis].terms.items():
            exp_value += Simulator.get_expectation_value_from_frequencies_oneterm(term, freqs) * coef

    return exp_value



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simulator class, wrapping around the various simulators and abstracting their
differences from the user. Able to run noiseless and noisy simulations,
leveraging the capabilities of different backends, quantum or classical.

If the user provides a noise model, then a noisy simulation is run with n_shots
shots. If the user only provides n_shots, a noiseless simulation is run, drawing
the desired amount of shots. If the target backend has access to the statevector
representing the quantum state, we leverage any kind of emulation available to
reduce runtime (directly generating shot values from final statevector etc) If
the quantum circuit contains a MEASURE instruction, it is assumed to simulate a
mixed-state and the simulation will be carried by simulating individual shots
(e.g a number of shots is required).

Some backends may only support a subset of the above. This information is
contained in a separate data-structure.
"""

import os
import math
from collections import Counter

import numpy as np
from scipy import stats
from bitarray import bitarray
from openfermion.ops import QubitOperator

from tangelo.helpers.utils import default_simulator
from tangelo.linq import Gate, Circuit
from tangelo.linq.helpers.circuits.measurement_basis import measurement_basis_gates
import tangelo.linq.translator as translator


# Data-structure showing what functionalities are supported by the backend, in this package
backend_info = dict()
backend_info["qiskit"] = {"statevector_available": True, "statevector_order": "msq_first", "noisy_simulation": True}
backend_info["qulacs"] = {"statevector_available": True, "statevector_order": "msq_first", "noisy_simulation": True}
backend_info["cirq"] = {"statevector_available": True, "statevector_order": "lsq_first", "noisy_simulation": True}
backend_info["qdk"] = {"statevector_available": False, "statevector_order": None, "noisy_simulation": False}


class Simulator:

    def __init__(self, target=default_simulator, n_shots=None, noise_model=None):
        """Instantiate Simulator object.

        Args:
            target (str): One of the available target backends (quantum or
                classical). The default simulator is qulacs if found in user's
                environment, otherwise cirq.
            n_shots (int): Number of shots if using a shot-based simulator.
            noise_model: A noise model object assumed to be in the format
                expected from the target backend.
        """
        self._source = "abstract"
        self._target = target if target else default_simulator
        self._current_state = None
        self._noise_model = noise_model

        # Can be modified later by user as long as long as it retains the same type (ex: cannot change to/from None)
        self.n_shots = n_shots
        self.freq_threshold = 1e-10

        # Set additional attributes related to the target backend chosen by the user
        for k, v in backend_info[self._target].items():
            setattr(self, k, v)

        # Raise error if user attempts to pass a noise model to a backend not supporting noisy simulation
        if self._noise_model and not self.noisy_simulation:
            raise ValueError("Target backend does not support noise models.")

        # Raise error if the number of shots has not been passed for a noisy simulation or if statevector unavailable
        if not self.n_shots and (not self.statevector_available or self._noise_model):
            raise ValueError("A number of shots needs to be specified.")

    def simulate(self, source_circuit, return_statevector=False, initial_statevector=None):
        """Perform state preparation corresponding to the input circuit on the
        target backend, return the frequencies of the different observables, and
        either the statevector or None depending on the availability of the
        statevector and if return_statevector is set to True. For the
        statevector backends supporting it, an initial statevector can be
        provided to initialize the quantum state without simulating all the
        equivalent gates.

        Args:
            source_circuit: a circuit in the abstract format to be translated
                for the target backend.
            return_statevector(bool): option to return the statevector as well,
                if available.
            initial_statevector(list/array) : A valid statevector in the format
                supported by the target backend.

        Returns:
            dict: A dictionary mapping multi-qubit states to their corresponding
                frequency.
            numpy.array: The statevector, if available for the target backend
                and requested by the user (if not, set to None).
        """

        if source_circuit.is_mixed_state and not self.n_shots:
            raise ValueError("Circuit contains MEASURE instruction, and is assumed to prepare a mixed state."
                             "Please set the Simulator.n_shots attribute to an appropriate value.")

        if source_circuit.width == 0:
            raise ValueError("Cannot simulate an empty circuit (e.g identity unitary) with unknown number of qubits.")

        # If the unitary is the identity (no gates) and no noise model, no need for simulation:
        # return all-zero state or sample from statevector
        if source_circuit.size == 0 and not self._noise_model:
            if initial_statevector is not None:
                statevector = initial_statevector
                frequencies = self._statevector_to_frequencies(initial_statevector)
            else:
                frequencies = {'0'*source_circuit.width: 1.0}
                statevector = np.zeros(2**source_circuit.width)
                statevector[0] = 1.0
            return (frequencies, statevector) if return_statevector else (frequencies, None)

        if self._target == "qulacs":
            import qulacs

            translated_circuit = translator.translate_qulacs(source_circuit, self._noise_model)

            # Initialize state on GPU if available and desired. Default to CPU otherwise.
            if ('QuantumStateGpu' in dir(qulacs)) and (int(os.getenv("QULACS_USE_GPU", 0)) != 0):
                state = qulacs.QuantumStateGpu(source_circuit.width)
            else:
                state = qulacs.QuantumState(source_circuit.width)
            if initial_statevector is not None:
                state.load(initial_statevector)

            if (source_circuit.is_mixed_state or self._noise_model):
                samples = list()
                for i in range(self.n_shots):
                    translated_circuit.update_quantum_state(state)
                    samples.append(state.sampling(1)[0])
                    if initial_statevector is not None:
                        state.load(initial_statevector)
                    else:
                        state.set_zero_state()
                python_statevector = None
            elif self.n_shots is not None:
                translated_circuit.update_quantum_state(state)
                python_statevector = np.array(state.get_vector()) if return_statevector else None
                samples = state.sampling(self.n_shots)
            else:
                translated_circuit.update_quantum_state(state)
                self._current_state = state
                python_statevector = state.get_vector()
                frequencies = self._statevector_to_frequencies(python_statevector)
                return (frequencies, np.array(python_statevector)) if return_statevector else (frequencies, None)

            frequencies = {self.__int_to_binstr(k, source_circuit.width): v / self.n_shots
                           for k, v in Counter(samples).items()}
            return (frequencies, python_statevector)

        elif self._target == "qiskit":
            import qiskit

            translated_circuit = translator.translate_qiskit(source_circuit)

            # If requested, set initial state
            if initial_statevector is not None:
                if self._noise_model:
                    raise ValueError("Cannot load an initial state if using a noise model, with Qiskit")
                else:
                    n_qubits = int(math.log2(len(initial_statevector)))
                    initial_state_circuit = qiskit.QuantumCircuit(n_qubits, n_qubits)
                    initial_state_circuit.initialize(initial_statevector, list(range(n_qubits)))
                    translated_circuit = initial_state_circuit.compose(translated_circuit)

            # Drawing individual shots with the qasm simulator, for noisy simulation or simulating mixed states
            if self._noise_model or source_circuit.is_mixed_state:
                from tangelo.linq.noisy_simulation.noise_models import get_qiskit_noise_model

                meas_range = range(source_circuit.width)
                translated_circuit.measure(meas_range, meas_range)
                return_statevector = False
                backend = qiskit.Aer.get_backend("aer_simulator")

                qiskit_noise_model = get_qiskit_noise_model(self._noise_model) if self._noise_model else None
                opt_level = 0 if self._noise_model else None

                job_sim = qiskit.execute(translated_circuit, backend, noise_model=qiskit_noise_model,
                                         shots=self.n_shots, basis_gates=None, optimization_level=opt_level)
                sim_results = job_sim.result()
                frequencies = {state[::-1]: count/self.n_shots for state, count in sim_results.get_counts(0).items()}

            # Noiseless simulation using the statevector simulator otherwise
            else:
                backend = qiskit.Aer.get_backend("aer_simulator", method='statevector')
                translated_circuit = qiskit.transpile(translated_circuit, backend)
                translated_circuit.save_statevector()
                sim_results = backend.run(translated_circuit).result()
                self._current_state = sim_results.get_statevector(translated_circuit)
                frequencies = self._statevector_to_frequencies(self._current_state)

            return (frequencies, np.array(sim_results.get_statevector())) if return_statevector else (frequencies, None)

        elif self._target == "qdk":

            translated_circuit = translator.translate_qsharp(source_circuit)
            with open('tmp_circuit.qs', 'w+') as f_out:
                f_out.write(translated_circuit)

            # Compile, import and call Q# operation to compute frequencies. Only import qsharp module if qdk is running
            # TODO: A try block to catch an exception at compile time, for Q#? Probably as an ImportError.
            import qsharp
            qsharp.reload()
            from MyNamespace import EstimateFrequencies
            frequencies_list = EstimateFrequencies.simulate(nQubits=source_circuit.width, nShots=self.n_shots)
            print("Q# frequency estimation with {0} shots: \n {1}".format(self.n_shots, frequencies_list))

            # Convert Q# output to frequency dictionary, apply threshold
            frequencies = {bin(i).split('b')[-1]: frequencies_list[i] for i, freq in enumerate(frequencies_list)}
            frequencies = {("0"*(source_circuit.width-len(k))+k)[::-1]: v for k, v in frequencies.items()
                           if v > self.freq_threshold}
            return (frequencies, None)

        elif self._target == "cirq":
            import cirq

            translated_circuit = translator.translate_cirq(source_circuit, self._noise_model)

            if source_circuit.is_mixed_state or self._noise_model:
                # Only DensityMatrixSimulator handles noise well, can use Simulator but it is slower
                cirq_simulator = cirq.DensityMatrixSimulator(dtype=np.complex128)
            else:
                cirq_simulator = cirq.Simulator(dtype=np.complex128)

            # If requested, set initial state
            cirq_initial_statevector = initial_statevector if initial_statevector is not None else 0

            # Calculate final density matrix and sample from that for noisy simulation or simulating mixed states
            if self._noise_model or source_circuit.is_mixed_state:
                # cirq.dephase_measurements changes measurement gates to Krauss operators so simulators
                # can be called once and density matrix sampled repeatedly.
                translated_circuit = cirq.dephase_measurements(translated_circuit)
                sim = cirq_simulator.simulate(translated_circuit, initial_state=cirq_initial_statevector)
                self._current_state = sim.final_density_matrix
                indices = list(range(source_circuit.width))
                isamples = cirq.sample_density_matrix(sim.final_density_matrix, indices, repetitions=self.n_shots)
                samples = [''.join([str(int(q))for q in isamples[i]]) for i in range(self.n_shots)]

                frequencies = {k: v / self.n_shots for k, v in Counter(samples).items()}
            # Noiseless simulation using the statevector simulator otherwise
            else:
                job_sim = cirq_simulator.simulate(translated_circuit, initial_state=cirq_initial_statevector)
                self._current_state = job_sim.final_state_vector
                frequencies = self._statevector_to_frequencies(self._current_state)

            return (frequencies, np.array(self._current_state)) if return_statevector else (frequencies, None)

    def get_expectation_value(self, qubit_operator, state_prep_circuit, initial_statevector=None):
        r"""Take as input a qubit operator H and a quantum circuit preparing a
        state |\psi>. Return the expectation value <\psi | H | \psi>.

        In the case of a noiseless simulation, if the target backend exposes the
        statevector then it is used directly to compute expectation values, or
        draw samples if required. In the case of a noisy simulator, or if the
        statevector is not available on the target backend, individual shots
        must be run and the workflow is akin to what we would expect from an
        actual QPU.

        Args:
            qubit_operator(openfermion-style QubitOperator class): a qubit
                operator.
            state_prep_circuit: an abstract circuit used for state preparation.

        Returns:
            complex: The expectation value of this operator with regards to the
                state preparation.
        """
        # Check if simulator supports statevector
        if initial_statevector is not None and not self.statevector_available:
            raise ValueError(f'Statevector not supported in {self._target}')

        # Check that qubit operator does not operate on qubits beyond circuit size.
        # Keep track if coefficients are real or not
        are_coefficients_real = True
        for term, coef in qubit_operator.terms.items():
            if state_prep_circuit.width < len(term):
                raise ValueError(f'Term {term} requires more qubits than the circuit contains ({state_prep_circuit.width})')
            if type(coef) in {complex, np.complex64, np.complex128}:
                are_coefficients_real = False

        # If the underlying operator is hermitian, expectation value is real and can be computed right away
        if are_coefficients_real:
            if self._noise_model or not self.statevector_available \
                    or state_prep_circuit.is_mixed_state or state_prep_circuit.size == 0:
                return self._get_expectation_value_from_frequencies(qubit_operator, state_prep_circuit, initial_statevector=initial_statevector)
            elif self.statevector_available:
                return self._get_expectation_value_from_statevector(qubit_operator, state_prep_circuit, initial_statevector=initial_statevector)

        # Else, separate the operator into 2 hermitian operators, use linearity and call this function twice
        else:
            qb_op_real, qb_op_imag = QubitOperator(), QubitOperator()
            for term, coef in qubit_operator.terms.items():
                qb_op_real.terms[term], qb_op_imag.terms[term] = coef.real, coef.imag
            qb_op_real.compress()
            qb_op_imag.compress()
            exp_real = self.get_expectation_value(qb_op_real, state_prep_circuit, initial_statevector=initial_statevector)
            exp_imag = self.get_expectation_value(qb_op_imag, state_prep_circuit, initial_statevector=initial_statevector)
            return exp_real if (exp_imag == 0.) else exp_real + 1.0j * exp_imag

    def _get_expectation_value_from_statevector(self, qubit_operator, state_prep_circuit, initial_statevector=None):
        r"""Take as input a qubit operator H and a state preparation returning a
        ket |\psi>. Return the expectation value <\psi | H | \psi>, computed
        without drawing samples (statevector only). Users should not be calling
        this function directly, please call "get_expectation_value" instead.

        Args:
            qubit_operator(openfermion-style QubitOperator class): a qubit
                operator.
            state_prep_circuit: an abstract circuit used for state preparation
                (only pure states).

        Returns:
            complex: The expectation value of this operator with regards to the
                state preparation.
        """
        n_qubits = state_prep_circuit.width

        expectation_value = 0.
        prepared_frequencies, prepared_state = self.simulate(state_prep_circuit, return_statevector=True, initial_statevector=initial_statevector)

        # Use fast built-in qulacs expectation value function if possible
        if self._target == "qulacs" and not self.n_shots:
            import qulacs

            # Note: This section previously used qulacs.quantum_operator.create_quantum_operator_from_openfermion_text but was changed
            # due to a memory leak. We can re-evaluate the implementation if/when Issue #303 (https://github.com/qulacs/qulacs/issues/303)
            # is fixed.
            operator = qulacs.Observable(n_qubits)
            for term, coef in qubit_operator.terms.items():
                pauli_string = "".join(f" {op} {qu}" for qu, op in term)
                operator.add_operator(coef, pauli_string)
            return operator.get_expectation_value(self._current_state).real

        # Use cirq built-in expectation_from_state_vector/epectation_from_density_matrix
        # noise model would require
        if self._target == "cirq" and not self.n_shots:
            import cirq

            GATE_CIRQ = translator.get_cirq_gates()
            qubit_labels = cirq.LineQubit.range(n_qubits)
            qubit_map = {q: i for i, q in enumerate(qubit_labels)}
            paulisum = 0.*cirq.PauliString(cirq.I(qubit_labels[0]))
            for term, coef in qubit_operator.terms.items():
                pauli_list = [GATE_CIRQ[pauli](qubit_labels[index]) for index, pauli in term]
                paulisum += cirq.PauliString(pauli_list, coefficient=coef)
            if self._noise_model:
                exp_value = paulisum.expectation_from_density_matrix(prepared_state, qubit_map)
            else:
                exp_value = paulisum.expectation_from_state_vector(prepared_state, qubit_map)
            return np.real(exp_value)

        # Otherwise, use generic statevector expectation value
        for term, coef in qubit_operator.terms.items():

            if len(term) > n_qubits:  # Cannot have a qubit index beyond circuit size
                raise ValueError(f"Size of operator {qubit_operator} beyond circuit width ({n_qubits} qubits)")
            elif not term:  # Empty term: no simulation needed
                expectation_value += coef
                continue

            if not self.n_shots:
                # Directly simulate and compute expectation value using statevector
                pauli_circuit = Circuit([Gate(pauli, index) for index, pauli in term], n_qubits=n_qubits)
                _, pauli_state = self.simulate(pauli_circuit, return_statevector=True, initial_statevector=prepared_state)

                delta = np.dot(pauli_state.real, prepared_state.real) + np.dot(pauli_state.imag, prepared_state.imag)
                expectation_value += coef * delta

            else:
                # Run simulation with statevector but compute expectation value with samples directly drawn from it
                basis_circuit = Circuit(measurement_basis_gates(term), n_qubits=state_prep_circuit.width)
                if basis_circuit.size > 0:
                    frequencies, _ = self.simulate(basis_circuit, initial_statevector=prepared_state)
                else:
                    frequencies = prepared_frequencies
                expectation_term = self.get_expectation_value_from_frequencies_oneterm(term, frequencies)
                expectation_value += coef * expectation_term

        return expectation_value

    def _get_expectation_value_from_frequencies(self, qubit_operator, state_prep_circuit, initial_statevector=None):
        r"""Take as input a qubit operator H and a state preparation returning a
        ket |\psi>. Return the expectation value <\psi | H | \psi> computed
        using the frequencies of observable states.

        Args:
            qubit_operator(openfermion-style QubitOperator class): a qubit
                operator.
            state_prep_circuit: an abstract circuit used for state preparation.

        Returns:
            complex: The expectation value of this operator with regards to the
                state preparation.
        """
        n_qubits = state_prep_circuit.width
        if not self.statevector_available or state_prep_circuit.is_mixed_state or self._noise_model:
            initial_circuit = state_prep_circuit
            if initial_statevector is not None and not self.statevector_available:
                raise ValueError(f'Backend {self._target} does not support statevectors')
            else:
                updated_statevector = initial_statevector
        else:
            initial_circuit = Circuit(n_qubits=n_qubits)
            _, updated_statevector = self.simulate(state_prep_circuit, return_statevector=True, initial_statevector=initial_statevector)

        expectation_value = 0.
        for term, coef in qubit_operator.terms.items():

            if len(term) > n_qubits:
                raise ValueError(f"Size of operator {qubit_operator} beyond circuit width ({n_qubits} qubits)")
            elif not term:  # Empty term: no simulation needed
                expectation_value += coef
                continue

            basis_circuit = Circuit(measurement_basis_gates(term))
            full_circuit = initial_circuit + basis_circuit if (basis_circuit.size > 0) else initial_circuit
            frequencies, _ = self.simulate(full_circuit, initial_statevector=updated_statevector)
            expectation_term = self.get_expectation_value_from_frequencies_oneterm(term, frequencies)
            expectation_value += coef * expectation_term

        return expectation_value

    @staticmethod
    def get_expectation_value_from_frequencies_oneterm(term, frequencies):
        """Return the expectation value of a single-term qubit-operator, given
        the result of a state-preparation.

        Args:
            term(openfermion-style QubitOperator object): a qubit operator, with
                only a single term.
            frequencies(dict): histogram of frequencies of measurements (assumed
                to be in lsq-first format).

        Returns:
            complex: The expectation value of this operator with regards to the
                state preparation.
        """

        if not frequencies.keys():
            return ValueError("Must pass a non-empty dictionary of frequencies.")
        n_qubits = len(list(frequencies.keys())[0])

        # Get term mask
        mask = ["0"] * n_qubits
        for index, op in term:
            mask[index] = "1"
        mask = "".join(mask)

        # Compute expectation value of the term
        expectation_term = 0.
        for basis_state, freq in frequencies.items():
            # Compute sample value using state_binstr and term mask, update term expectation value
            sample = (-1) ** ((bitarray(mask) & bitarray(basis_state)).to01().count("1") % 2)
            expectation_term += sample * freq

        return expectation_term

    def _statevector_to_frequencies(self, statevector):
        """For a given statevector representing the quantum state of a qubit
        register, returns a sparse histogram of the probabilities in the
        least-significant-qubit (lsq) -first order. e.g the string '100' means
        qubit 0 measured in basis state |1>, and qubit 1 & 2 both measured in
        state |0>.

        Args:
            statevector(list or ndarray(complex)): an iterable 1D data-structure
                containing the amplitudes.

        Returns:
            dict: A dictionary whose keys are bitstrings representing the
                multi-qubit states with the least significant qubit first (e.g
                '100' means qubit 0 in state |1>, and qubit 1 and 2 in state
                |0>), and the associated value is the corresponding frequency.
                Unless threshold=0., this dictionary will be sparse.
        """

        n_qubits = int(math.log2(len(statevector)))
        frequencies = dict()
        for i, amplitude in enumerate(statevector):
            frequency = abs(amplitude)**2
            if (frequency - self.freq_threshold) >= 0.:
                frequencies[self.__int_to_binstr(i, n_qubits)] = frequency

        # If n_shots, has been specified, then draw that amount of samples from the distribution
        # and return empirical frequencies instead. Otherwise, return the exact frequencies
        if not self.n_shots:
            return frequencies
        else:
            xk, pk = [], []
            for k, v in frequencies.items():
                xk.append(int(k[::-1], 2))
                pk.append(frequencies[k])
            distr = stats.rv_discrete(name='distr', values=(np.array(xk), np.array(pk)))

            # Generate samples from distribution. Cut in chunks to ensure samples fit in memory, gradually accumulate
            chunk_size = 10**7
            n_chunks = self.n_shots // chunk_size
            freqs_shots = Counter()

            for i in range(n_chunks+1):
                this_chunk = self.n_shots % chunk_size if i == n_chunks else chunk_size
                samples = distr.rvs(size=this_chunk)
                freqs_shots += Counter(samples)
            freqs_shots = {self.__int_to_binstr_lsq(k, n_qubits): v / self.n_shots for k, v in freqs_shots.items()}
            return freqs_shots

    def __int_to_binstr(self, i, n_qubits):
        """Convert an integer into a bit string of size n_qubits, in the order
        specified for the state vector.
        """
        bs = bin(i).split('b')[-1]
        state_binstr = "0" * (n_qubits - len(bs)) + bs
        return state_binstr[::-1] if (self.statevector_order == "msq_first") else state_binstr

    def __int_to_binstr_lsq(self, i, n_qubits):
        """Convert an integer into a bit string of size n_qubits, in the
        least-significant qubit order.
        """
        bs = bin(i).split('b')[-1]
        state_binstr = "0" * (n_qubits - len(bs)) + bs
        return state_binstr[::-1]



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the qubit coupled cluster (QCC) ansatz class. The
motivation behind this ansatz is to provide an improved alternative to
classical unitary coupled cluster for describing the electron correlation of
molecular systems. This implementation is based on Ref. 1, where the ansaztz
takes the form of a variational product state built directly from a set of
parameterized exponentiated qubit operators. A qubit operator is selected for
the ansatz based on an energy gradient criterion that indicates its potential
contribution to variational lowering of the QCC energy. For chemical applications,
the quantum mean-field ansatz is used in conjunction with this ansatz to
describe an electronic wave function on a quantum computer. For more information
about this ansatz and its variations, see references below.

Refs:
    1. I. G. Ryabinkin, T.-C. Yen, S. N. Genin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2018, 14 (12), 6317-6326.
    2. I. G. Ryabinkin, R. A. Lang, S. N. Genin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2020, 16, 2, 1055–1063.
    3. R. A. Lang, I. G. Ryabinkin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2021, 17, 1, 66–78.
"""

import warnings

import numpy as np

from tangelo.toolboxes.qubit_mappings.mapping_transform import get_qubit_number,\
                                                               fermion_to_qubit_mapping
from tangelo.linq import Circuit
from tangelo.toolboxes.ansatz_generator.ansatz import Ansatz
from tangelo.toolboxes.ansatz_generator.ansatz_utils import exp_pauliword_to_gates
from tangelo.toolboxes.ansatz_generator._qubit_mf import init_qmf_from_hf, get_qmf_circuit, purify_qmf_state
from tangelo.toolboxes.ansatz_generator._qubit_cc import build_qcc_qubit_op, construct_dis


class QCC(Ansatz):
    """This class implements the QCC ansatz. Closed-shell and restricted open-shell QCC are
    supported. While the form of the QCC ansatz is the same for either variation, the underlying
    fermionic mean-field state is treated differently depending on the spin. Closed-shell
    or restricted open-shell QCC implies that spin = 0 or spin != 0 and the fermionic mean-field
    state is obtained using a RHF or ROHF Hamiltonian, respectively.

    Args:
        molecule (SecondQuantizedMolecule): The molecular system.
        mapping (str): One of the supported qubit mapping identifiers. Default, "jw".
        up_then_down (bool): Change basis ordering putting all spin-up orbitals first,
            followed by all spin-down. Default, False.
        dis (list of QubitOperator): The direct interaction set (DIS) of generators for the
            QCC ansatz. Default, None.
        qmf_circuit (Circuit): An instance of tangelo.linq Circuit class implementing a QMF state
            circuit. If passed from the QMF ansatz class, parameters are variational.
            If None, one is created with QMF parameters that are not variational. Default, None.
        qmf_var_params (list or numpy array of float): QMF variational parameter set.
            If None, the values are determined using a Hartree-Fock reference state. Default, None.
        qubit_ham (QubitOperator): Pass a qubit Hamiltonian to the QCC ansatz class and ignore
            the fermionic Hamiltonian in molecule. Default, None.
        deqcc_dtau_thresh (float): Threshold for |dEQCC/dtau| so that a candidate group is added
            to the DIS if |dEQCC/dtau| >= deqcc_dtau_thresh for a generator. Default, 1e-3 a.u.
        qcc_tau_guess (float): The initial guess for all QCC variational parameters.
            Default, 1e-2 a.u.
        max_qcc_gens (int or None): Maximum number of generators allowed in the ansatz. If None,
            one generator from each DIS group is selected. If int, then min(|DIS|, max_qcc_gens)
            generators are selected in order of decreasing |dEQCC/dtau|. Default, None.
    """

    def __init__(self, molecule, mapping="jw", up_then_down=False, dis=None,
                 qmf_circuit=None, qmf_var_params=None, qubit_ham=None, qcc_tau_guess=1e-2,
                 deqcc_dtau_thresh=1e-3, max_qcc_gens=None):

        if not molecule:
            raise ValueError("An instance of SecondQuantizedMolecule is required for initializing "
                             "the self.__class__.__name__ ansatz class.")
        self.molecule = molecule
        self.mapping = mapping
        self.up_then_down = up_then_down
        if self.mapping.lower() == "jw" and not self.up_then_down:
            warnings.warn("Spin-orbital ordering shifted to all spin-up first then down to "
                          "ensure efficient generator screening for the Jordan-Wigner mapping "
                          "with the self.__class__.__name__ ansatz.", RuntimeWarning)
            self.up_then_down = True

        self.n_spinorbitals = self.molecule.n_active_sos
        if self.n_spinorbitals % 2 != 0:
            raise ValueError("The total number of spin-orbitals should be even.")

        self.spin = molecule.spin
        self.fermi_ham = self.molecule.fermionic_hamiltonian
        self.n_electrons = self.molecule.n_electrons
        self.n_qubits = get_qubit_number(self.mapping, self.n_spinorbitals)

        self.qubit_ham = qubit_ham
        if not qubit_ham:
            self.fermi_ham = self.molecule.fermionic_hamiltonian
            self.qubit_ham = fermion_to_qubit_mapping(self.fermi_ham, self.mapping,
                                                      self.n_spinorbitals, self.n_electrons,
                                                      self.up_then_down, self.spin)

        self.qmf_var_params = qmf_var_params
        if not self.qmf_var_params:
            self.qmf_var_params = init_qmf_from_hf(self.n_spinorbitals, self.n_electrons,
                                                   self.mapping, self.up_then_down, self.spin)
        elif isinstance(self.qmf_var_params, list):
            self.qmf_var_params = np.array(self.qmf_var_params)
        if self.qmf_var_params.size != 2 * self.n_qubits:
            raise ValueError("The number of QMF variational parameters must be 2 * n_qubits.")
        self.n_qmf_params = 2 * self.n_qubits
        self.qmf_circuit = qmf_circuit

        self.dis = dis
        self.qcc_tau_guess = qcc_tau_guess
        self.deqcc_dtau_thresh = deqcc_dtau_thresh
        self.max_qcc_gens = max_qcc_gens

        # Build the DIS or specify a list of generators; updates the number of QCC parameters
        self._get_qcc_generators()
        self.n_var_params = self.n_qmf_params + self.n_qcc_params

        # Supported reference state initialization
        self.supported_reference_state = {"HF"}
        # Supported var param initialization
        self.supported_initial_var_params = {"qmf_state", "qcc_tau_guess", "random"}

        # Default starting parameters for initialization
        self.pauli_to_angles_mapping = {}
        self.default_reference_state = "HF"
        self.var_params_default = "qcc_tau_guess"
        self.var_params = None
        self.qcc_circuit = None
        self.circuit = None

    def set_var_params(self, var_params=None):
        """Set values for variational parameters, such as zeros or floats,
        providing some keywords for users, and also supporting direct user input
        (list or numpy array). Return the parameters so that workflows such as VQE can
        retrieve these values. """

        if var_params is None:
            var_params = self.var_params_default

        if isinstance(var_params, str):
            var_params = var_params.lower()
            if var_params not in self.supported_initial_var_params:
                raise ValueError(f"Supported keywords for initializing variational parameters: "
                                 f"{self.supported_initial_var_params}")
            # Initialize the QCC wave function as |QCC> = |QMF>
            if var_params == "qmf_state":
                initial_var_params = np.zeros((self.n_qcc_params,), dtype=float)
            # Initialize all tau parameters to the same value specified by self.qcc_tau_guess
            elif var_params == "qcc_tau_guess":
                initial_var_params = self.qcc_tau_guess * np.ones((self.n_qcc_params,))
            # Initialize tau parameters randomly over the domain [0., 2 pi)
            elif var_params == "random":
                initial_var_params = 2. * np.pi * np.random.random((self.n_qcc_params,))
            # Insert the QMF variational parameters at the beginning.
            initial_var_params = np.concatenate((self.qmf_var_params, initial_var_params))
        else:
            initial_var_params = np.array(var_params)
        self.var_params = initial_var_params
        if initial_var_params.size != self.n_var_params:
            raise ValueError(f"Expected {self.n_var_params} variational parameters but "
                             f"received {initial_var_params.size}.")
        return initial_var_params

    def prepare_reference_state(self):
        """Returns circuit preparing the reference state of the ansatz (e.g prepare reference
        wavefunction with HF, multi-reference state, etc). These preparations must be consistent
        with the transform used to obtain the qubit operator. """

        if self.default_reference_state not in self.supported_reference_state:
            raise ValueError(f"Only supported reference state methods are: "
                             f"{self.supported_reference_state}.")
        if self.default_reference_state == "HF":
            reference_state_circuit = get_qmf_circuit(self.qmf_var_params, True)
        return reference_state_circuit

    def build_circuit(self, var_params=None):
        """Build and return the quantum circuit implementing the state preparation ansatz
         (with currently specified initial_state and var_params). """

        # Build the DIS or specify a list of generators; updates the number of QCC parameters
        self._get_qcc_generators()
        self.n_var_params = self.n_qmf_params + self.n_qcc_params

        # Get the variational parameters needed for the QCC unitary operator and circuit
        if var_params is not None:
            self.set_var_params(var_params)
        elif self.var_params is None:
            self.set_var_params()

        # Build a QMF state preparation circuit
        if self.qmf_circuit is None:
            self.qmf_circuit = self.prepare_reference_state()

        # Build the QCC ansatz qubit operator
        qubit_op = build_qcc_qubit_op(self.dis, self.var_params[2*self.n_qubits:])

        # Obtain quantum circuit through trivial trotterization of the qubit operator
        # Track the order in which pauli words have been visited for fast parameter updates
        pauli_words_gates = []
        pauli_words = sorted(qubit_op.terms.items(), key=lambda x: len(x[0]))
        for i, (pauli_word, coef) in enumerate(pauli_words):
            pauli_words_gates += exp_pauliword_to_gates(pauli_word, coef)
            self.pauli_to_angles_mapping[pauli_word] = i
        self.qcc_circuit = Circuit(pauli_words_gates)
        self.circuit = self.qmf_circuit + self.qcc_circuit if self.qmf_circuit.size != 0\
                       else self.qcc_circuit

    def update_var_params(self, var_params):
        """Shortcut: set value of variational parameters in the existing ansatz circuit member.
        Preferable to rebuilding your circuit from scratch, which can be an involved process.
        """

        # Update the QCC variational parameters
        self.set_var_params(var_params)

        # Build the QCC ansatz qubit operator
        qubit_op = build_qcc_qubit_op(self.dis, self.var_params[2*self.n_qubits:])

        # If qubit_op terms have changed, rebuild circuit
        if set(self.pauli_to_angles_mapping.keys()) != set(qubit_op.terms.keys()):
            self.build_circuit(var_params)
        # Otherwise update variational gates directly
        else:
            for pauli_word, coef in qubit_op.terms.items():
                gate_index = self.pauli_to_angles_mapping[pauli_word]
                gate_param = 2. * coef if coef >= 0. else 4 * np.pi + 2 * coef
                self.qcc_circuit._variational_gates[gate_index].parameter = gate_param
            self.circuit = self.qmf_circuit + self.qcc_circuit if self.qmf_circuit.size != 0\
                           else self.qcc_circuit

    def _get_qcc_generators(self):
        """ Prepares the QCC ansatz by purifying the QMF state, constructing the DIS,
        and selecting representative generators from the top candidate DIS groups. """

        if not self.dis:
            pure_var_params = purify_qmf_state(self.qmf_var_params, self.n_spinorbitals, self.n_electrons,
                                               self.mapping, self.up_then_down, self.spin)
            self.dis = construct_dis(self.qubit_ham, pure_var_params, self.deqcc_dtau_thresh)
            if self.max_qcc_gens:
                self.n_qcc_params = min(len(self.dis), self.max_qcc_gens)
                del self.dis[self.n_qcc_params:]
            else:
                self.n_qcc_params = len(self.dis)
        else:
            self.n_qcc_params = len(self.dis)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Construct the orbital list for the fragments.

Construct the orbital list, showing how much orbitals to be included for each
fragment, from the atom list, showing how many atoms included in each fragment.
"""


def dmet_fragment_constructor(mol, atom_list, number_fragment):
    """Construct orbital list.

    Make a list of number of orbitals for each fragment while obtaining the list
    if we consider combining fragments.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate (The full molecule).
        atom_list (list): The atom list for each fragment (int).
        number_fragment (list): Number of element in atom list per fragment
            (int).

    Returns:
        list: The number of orbitals for each fragment (int).
        list: List of lists of the minimum and maximum orbital label for each
            fragment (int).
        list: The new atom list for each fragment (int).
    """

    # Make a new atom list based on how many fragments for DMET calculation
    if number_fragment == 0:
        atom_list2 = atom_list
    else:
        # Calculate the number of DMET calculations
        number_new_fragment = int(len(atom_list)/(number_fragment+1))  # number of DMET calulation per loop
        atom_list2 = []

        # Define the number of atoms per DMET calculation
        for i in range(number_new_fragment):
            num = 0
            for j in range(number_fragment + 1):
                k = (number_fragment+1)*i+j
                num += atom_list[k]
            atom_list2.append(num)

    # Initialize the list of the number of orbitals
    orb_list = []
    orb_list2 = []
    isum = 0
    isum2 = -1
    iorb = 0
    jorb = 0

    # Calculate the number of orbitals for each atom
    for i in atom_list2:
        itemp = 0
        isum2 += i
        for total_basis in mol.spheric_labels():
            item = total_basis.split()
            item0 = int(item[0])
            if (isum <= item0 <= isum2):
                itemp += 1
        isum += i
        jorb += itemp
        orb_list.append(itemp)
        orb_list2.append([iorb, jorb])
        iorb += itemp

    return orb_list, orb_list2, atom_list2



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os

from openfermion import load_operator
from openfermion.ops import QubitOperator

from tangelo.linq import translator, Simulator, Circuit
from tangelo.helpers import measurement_basis_gates
from tangelo.toolboxes.measurements import group_qwc, exp_value_from_measurement_bases, \
    check_bases_commute_qwc, map_measurements_qwc

path_data = os.path.dirname(os.path.abspath(__file__)) + '/data'


class TermsGroupingTest(unittest.TestCase):

    def test_qubitwise_commutativity(self):
        """ Tests for check_bases_commute_qwc function. """

        I0 = ()
        Z1 = ((1, 'Z'),)
        X2 = ((2, 'X'),)
        X0Z1 = ((0, 'X'), (1, 'Z'))
        Z0Z1 = ((0, 'Z'), (1, 'Z'))

        self.assertTrue(check_bases_commute_qwc(I0, Z1))
        self.assertTrue(check_bases_commute_qwc(X2, Z1))
        self.assertTrue(check_bases_commute_qwc(Z1, X0Z1))
        self.assertTrue(check_bases_commute_qwc(Z1, Z0Z1))
        self.assertFalse(check_bases_commute_qwc(Z0Z1, X0Z1))

    def test_mmap_qwc_H2(self):
        """ From a partitioned Hamiltonian, build reverse dictionary of measurements """

        qb_ham = load_operator("mol_H2_qubitham.data", data_directory=path_data, plain_text=True)
        partitioned_ham = group_qwc(qb_ham)
        res = map_measurements_qwc(partitioned_ham)

        for k, v in res.items():
            if k in {((0, 'X'), (1, 'X'), (2, 'Y'), (3, 'Y')), ((0, 'X'), (1, 'Y'), (2, 'Y'), (3, 'X')),
                     ((0, 'Y'), (1, 'X'), (2, 'X'), (3, 'Y')), ((0, 'Y'), (1, 'Y'), (2, 'X'), (3, 'X'))}:
                self.assertTrue(v[0] == k)
            else:
                self.assertTrue(v[0] == ((0, 'Z'), (1, 'Z'), (2, 'Z'), (3, 'Z')))

    def test_mmap_qwc_H10frag(self):
        """ From a partitioned Hamiltonian, build reverse dictionary of measurements """

        partitioned_ham = {
            ((0, 'X'), (1, 'X')): 0.25 * QubitOperator('X0 X1'),
            ((0, 'X'), (1, 'Z')): 0.25 * (QubitOperator('X0') + QubitOperator('X0 Z1') + QubitOperator('Z1')),
            ((0, 'Y'), (1, 'Y')): -0.25 * QubitOperator('Y0 Y1'),
            ((0, 'Z'), (1, 'X')): 0.25 * (QubitOperator('Z0') + QubitOperator('Z0 X1') + QubitOperator('X1')),
            ((0, 'Z'), (1, 'Z')): 0.25 * QubitOperator('Z0 Z1')
        }
        res = map_measurements_qwc(partitioned_ham)

        ref_H10frag = {((0, 'X'), (1, 'X')): [((0, 'X'), (1, 'X'))],
                       ((0, 'X'),): [((0, 'X'), (1, 'X')), ((0, 'X'), (1, 'Z'))],
                       ((0, 'X'), (1, 'Z')): [((0, 'X'), (1, 'Z'))],
                       ((1, 'Z'),): [((0, 'X'), (1, 'Z')), ((0, 'Z'), (1, 'Z'))],
                       ((0, 'Y'), (1, 'Y')): [((0, 'Y'), (1, 'Y'))],
                       ((0, 'Z'),): [((0, 'Z'), (1, 'X')), ((0, 'Z'), (1, 'Z'))],
                       ((0, 'Z'), (1, 'X')): [((0, 'Z'), (1, 'X'))],
                       ((1, 'X'),): [((0, 'X'), (1, 'X')), ((0, 'Z'), (1, 'X'))],
                       ((0, 'Z'), (1, 'Z')): [((0, 'Z'), (1, 'Z'))]}

        self.assertDictEqual(res, ref_H10frag)

    def test_qubitwise_commutativity_of_H2(self):
        """ The JW Pauli hamiltonian of H2 at optimal geometry is a 15-term operator. Using qubitwise-commutativity,
        it is possible to get the expectation value of all 15 terms by only performing measurements in 5 distinct
        measurement bases. This test first verifies the 5 measurement bases have been identified, and then derives
        the expectation value for the qubit Hamiltonian.
        """

        # Load qubit Hamiltonian
        qb_ham = load_operator("mol_H2_qubitham.data", data_directory=path_data, plain_text=True)

        # Group Hamiltonian terms using qubitwise commutativity
        grouped_ops = group_qwc(qb_ham, seed=0)

        # Load an optimized quantum circuit (UCCSD) to compute something meaningful in this test
        with open(f"{path_data}/H2_UCCSD.qasm", "r") as f:
            openqasm_circ = f.read()
        abs_circ = translator._translate_openqasm2abs(openqasm_circ)

        # Only simulate and measure the wavefunction in the required bases (simulator or QPU), store in dict.
        histograms = dict()
        sim = Simulator()
        for basis, sub_op in grouped_ops.items():
            full_circuit = abs_circ + Circuit(measurement_basis_gates(basis))
            histograms[basis], _ = sim.simulate(full_circuit)

        # Reconstruct exp value of initial input operator using the histograms corresponding to the suboperators
        exp_value = exp_value_from_measurement_bases(grouped_ops, histograms)
        self.assertAlmostEqual(exp_value, sim.get_expectation_value(qb_ham, abs_circ), places=8)

    def test_qubitwise_commutativity_of_H4(self):
        """ Estimating the energy of a rectangle configuration of H4, resulting in a 185-term qubit operator.
        Uses qubitwise-commutativity to identify the number of measurement bases needed (~60) to compute the
        expectation value of the full operator using a pre-loaded UCCSD H4 circuit.
        """

        # Load qubit Hamiltonian
        qb_ham = load_operator("mol_H4_qubitham.data", data_directory=path_data, plain_text=True)

        # Group Hamiltonian terms using qubitwise commutativity
        grouped_ops = group_qwc(qb_ham, seed=0)

        # Load an optimized quantum circuit (UCCSD) to compute something meaningful in this test
        with open(f"{path_data}/H4_UCCSD.qasm", "r") as f:
            openqasm_circ = f.read()
        abs_circ = translator._translate_openqasm2abs(openqasm_circ)

        # Only simulate and measure the wavefunction in the required bases (simulator or QPU), store in dict.
        histograms = dict()
        sim = Simulator()
        for basis, sub_op in grouped_ops.items():
            full_circuit = abs_circ + Circuit(measurement_basis_gates(basis))
            histograms[basis], _ = sim.simulate(full_circuit)

        # Reconstruct exp value of initial input operator using the histograms corresponding to the suboperators
        exp_value = exp_value_from_measurement_bases(grouped_ops, histograms)
        self.assertAlmostEqual(exp_value, sim.get_expectation_value(qb_ham, abs_circ), places=8)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import math

from openfermion import get_sparse_operator
import numpy as np
from scipy.linalg import expm

from tangelo.linq import Simulator, backend_info
from tangelo.helpers.utils import installed_backends
from tangelo.linq.helpers.circuits.statevector import StateVector
from tangelo.toolboxes.operators.operators import QubitOperator
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.ansatz_generator.ansatz_utils import get_qft_circuit
from tangelo.molecule_library import mol_H2_sto3g
from tangelo.toolboxes.circuits.lcu import get_lcu_circuit, get_truncated_taylor_series

# Test for both "cirq" and if available "qulacs". These have different orderings.
# qiskit is not currently supported because does not have multi controlled general gates.
backends = ["cirq", "qulacs"] if "qulacs" in installed_backends else ["cirq"]
# Initiate Simulator using cirq for phase estimation tests as it has the same ordering as openfermion
# and we are using an exact eigenvector for testing.
sim_cirq = Simulator("cirq")


class lcu_Test(unittest.TestCase):

    def test_get_truncated_taylor_series(self):
        """Test time-evolution of truncated Taylor series for different orders and times"""

        qu_op = fermion_to_qubit_mapping(mol_H2_sto3g.fermionic_hamiltonian, "scbk", mol_H2_sto3g.n_active_sos, mol_H2_sto3g.n_active_electrons,
                                         True, 0)
        n_qubits_qu_op = math.ceil(math.log2(len(qu_op.terms)))
        ham = get_sparse_operator(qu_op).toarray()
        _, vecs = np.linalg.eigh(ham)
        vec = (vecs[:, 0] + vecs[:, 1])/np.sqrt(2)

        time = 1.9
        exact = expm(-1j*ham*time)@vec

        for backend in backends:
            sim = Simulator(backend)
            statevector_order = backend_info[backend]["statevector_order"]
            sv = StateVector(vec, order=statevector_order)
            sv_circuit = sv.initializing_circuit()

            # Tested for up to k = 5 but 5 is slow due to needing 23 qubits to simulate.
            for k in [1, 2, 3, 4]:
                taylor_circuit = get_truncated_taylor_series(qu_op, k, time)
                _, v = sim.simulate(sv_circuit + taylor_circuit, return_statevector=True)
                len_ancilla = 2**(k+k*n_qubits_qu_op+1)
                v = v.reshape([4, len_ancilla])[:, 0] if statevector_order == "lsq_first" else v.reshape([len_ancilla, 4])[0, :]
                self.assertAlmostEqual(1, np.abs(v.conj().dot(exact)), delta=3.e-1**k)

        # Raise ValueError if Taylor series order is less than 1 or greater than 4 or imaginary coefficients in qubit operator
        self.assertRaises(ValueError, get_truncated_taylor_series, qu_op, 0, time)
        self.assertRaises(ValueError, get_truncated_taylor_series, qu_op * 1j, 2, time)

    def test_get_lcu_circuit(self):
        """Test time-evolution of truncated Taylor series for order k = 3 passing explicitly calculated qubit operator exponential"""

        qu_op = fermion_to_qubit_mapping(mol_H2_sto3g.fermionic_hamiltonian, "scbk", mol_H2_sto3g.n_active_sos, mol_H2_sto3g.n_active_electrons,
                                         True, 0)
        time = 0.5
        # Generate explicit qubit operator exponential
        exp_qu_op = 1 + -1j*qu_op*time + (-1j*qu_op*time)**2/2 + (-1j*qu_op*time)**3/6
        exp_qu_op.compress()
        n_qubits_qu_op = math.ceil(math.log2(len(exp_qu_op.terms)))
        ham = get_sparse_operator(qu_op).toarray()
        _, vecs = np.linalg.eigh(ham)
        vec = (vecs[:, 0] + vecs[:, 1])/np.sqrt(2)

        exact = expm(-1j*ham*time)@vec
        len_ancilla = 2**(n_qubits_qu_op)

        for backend in backends:
            sim = Simulator(backend)
            statevector_order = backend_info[backend]["statevector_order"]
            sv = StateVector(vec, order=statevector_order)
            sv_circuit = sv.initializing_circuit()

            taylor_circuit = get_lcu_circuit(exp_qu_op)
            _, v = sim.simulate(sv_circuit + taylor_circuit, return_statevector=True)
            v = v.reshape([4, len_ancilla])[:, 0] if statevector_order == "lsq_first" else v.reshape([len_ancilla, 4])[0, :]
            self.assertAlmostEqual(1, np.abs(v.conj().dot(exact)), delta=1.e-3)

        # Test return of ValueError if 1-norm is greater than 2.
        self.assertRaises(ValueError, get_lcu_circuit, exp_qu_op+5)

    def test_controlled_time_evolution_by_phase_estimation_for_get_truncated_taylor_series(self):
        """ Verify that the controlled time-evolution is correct by calculating the eigenvalue of an eigenstate through
        phase estimation.
        """

        # Generate qubit operator with state 9 having eigenvalue 0.25
        qu_op = (QubitOperator("X0 X1", 0.125) + QubitOperator("Y1 Y2", 0.125) + QubitOperator("Z2 Z3", 0.125)
                 + QubitOperator("", 0.125))

        ham_mat = get_sparse_operator(qu_op).toarray()
        _, wavefunction = np.linalg.eigh(ham_mat)

        # Kronecker product 13 qubits in the zero state to eigenvector 9 to account for ancilla qubits
        wave_9 = wavefunction[:, 9]
        for i in range(13):
            wave_9 = np.kron(wave_9, np.array([1, 0]))

        qubit_list = [16, 15, 14]

        pe_circuit = get_qft_circuit(qubit_list)
        for i, qubit in enumerate(qubit_list):
            pe_circuit += get_truncated_taylor_series(qu_op, 3, t=-(2*np.pi)*2**i, control=qubit)
        pe_circuit += get_qft_circuit(qubit_list, inverse=True)

        freqs, _ = sim_cirq.simulate(pe_circuit, initial_statevector=wave_9)

        # Trace out all but final 3 indices
        trace_freq = dict()
        for key, value in freqs.items():
            trace_freq[key[-3:]] = trace_freq.get(key[-3:], 0) + value

        # State 9 has eigenvalue 0.25 so return should be 010 (0*1/2 + 1*1/4 + 0*1/8)
        self.assertAlmostEqual(trace_freq["010"], 1.0, delta=1.e-4)

    def test_controlled_time_evolution_by_phase_estimation_for_get_lcu_circuit(self):
        """ Verify that the controlled time-evolution is correct by calculating the eigenvalue of an eigenstate through
        phase estimation.
        """

        # Generate qubit operator with state 9 having eigenvalue 0.25
        qu_op = (QubitOperator("X0 X1", 0.125) + QubitOperator("Y1 Y2", 0.125) + QubitOperator("Z2 Z3", 0.125)
                 + QubitOperator("", 0.125))

        ham_mat = get_sparse_operator(qu_op).toarray()
        _, wavefunction = np.linalg.eigh(ham_mat)

        # break time into 6 parts so 1-norm is less than 2. i.e. can use Oblivious Amplitude Amplification
        time = -2 * np.pi / 6

        # Generate explicit qubit operator exponential
        exp_qu_op = 1 + -1j*qu_op*time + (-1j*qu_op*time)**2/2 + (-1j*qu_op*time)**3/6

        num_terms = len(exp_qu_op.terms)
        n_extra_qubits = math.ceil(np.log2(num_terms))

        # Kronecker product n_extra_qubits + 3 qubits in the zero state to eigenvector 9
        wave_9 = wavefunction[:, 9]
        for i in range(n_extra_qubits+3):
            wave_9 = np.kron(wave_9, np.array([1, 0]))

        # Phase estimation register in reversed order
        qubit_list = list(reversed(range(4 + n_extra_qubits, 4 + n_extra_qubits + 3)))

        # Build phase estimation circuit and simulate
        pe_circuit = get_qft_circuit(qubit_list)
        for i, qubit in enumerate(qubit_list):
            pe_circuit += (6 * 2**i) * get_lcu_circuit(exp_qu_op, control=qubit)
        pe_circuit += get_qft_circuit(qubit_list, inverse=True)

        freqs, _ = sim_cirq.simulate(pe_circuit, initial_statevector=wave_9)

        # Trace out all but final 3 indices
        trace_freq = dict()
        for key, value in freqs.items():
            trace_freq[key[-3:]] = trace_freq.get(key[-3:], 0) + value

        # State 9 has eigenvalue 0.25 so return should be 010 (0*1/2 + 1*1/4 + 0*1/8)
        self.assertAlmostEqual(trace_freq["010"], 1.0, delta=1.e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that defines the Quantum Imaginary Time Algorithm (QITE)
"""
from copy import copy
import math

from openfermion import FermionOperator as ofFermionOperator
import numpy as np

from tangelo.toolboxes.ansatz_generator.ansatz_utils import trotterize
from tangelo.toolboxes.operators.operators import FermionOperator, QubitOperator
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.ansatz_generator._general_unitary_cc import uccgsd_generator as uccgsd_pool
from tangelo.toolboxes.operators import qubitop_to_qubitham
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit
from tangelo.linq import Circuit, Simulator


class QITESolver:
    """QITE class. This is an iterative algorithm that obtains a unitary operator
    that approximates the imaginary time evolution of an initial state.

    Attributes:
        molecule (SecondQuantizedMolecule): The molecular system.
        dt (float): The imaginary time step size
        min_de (float): Maximum energy change allowed before convergence.
        max_cycles (int): Maximum number of iterations of QITE.
        pool (func): Function that returns a list of FermionOperator. Each
            element represents an excitation/operator that has an effect on the
            total energy.
        pool_args (tuple) : The arguments for the pool function given as a
            tuple.
        qubit_mapping (str): One of the supported qubit mapping identifiers.
        qubit_hamiltonian (QubitOperator-like): Self-explanatory.
        up_then_down (bool): Spin orbitals ordering.
        n_spinorbitals (int): Self-explanatory.
        n_electrons (int): Self-explanatory.
        backend_options (dict): Backend options for the underlying QITE propagation
        verbose (bool): Flag for verbosity of QITE.
     """

    def __init__(self, opt_dict):

        default_backend_options = {"target": None, "n_shots": None, "noise_model": None}
        default_options = {"molecule": None,
                           "dt": 0.5, "max_cycles": 100,
                           "min_de": 1.e-7,
                           "pool": uccgsd_pool,
                           "pool_args": None,
                           "frozen_orbitals": "frozen_core",
                           "qubit_mapping": "jw",
                           "qubit_hamiltonian": None,
                           "up_then_down": False,
                           "n_spinorbitals": None,
                           "n_electrons": None,
                           "backend_options": default_backend_options,
                           "verbose": True}

        # Initialize with default values
        self.__dict__ = default_options
        # Overwrite default values with user-provided ones, if they correspond to a valid keyword
        for k, v in opt_dict.items():
            if k in default_options:
                setattr(self, k, v)
            else:
                raise KeyError(f"Keyword :: {k}, not available in {self.__class__.__name__}")

        # Raise error/warnings if input is not as expected. Only a single input
        # must be provided to avoid conflicts.
        if not (bool(self.molecule) ^ bool(self.qubit_hamiltonian)):
            raise ValueError("A molecule OR qubit Hamiltonian object must be provided when instantiating "
                             f"{self.__class__.__name__}.")

        if self.qubit_hamiltonian:
            if not (self.n_spinorbitals and self.n_electrons):
                raise ValueError("Expecting the number of spin-orbitals (n_spinorbitals) and the number of "
                                 "electrons (n_electrons) with a qubit_hamiltonian.")

        self.iteration = 0
        self.energies = list()

        self.circuit_list = list()
        self.final_energy = None
        self.final_circuit = None
        self.final_statevector = None

        self.backend = None

    def prepare_reference_state(self):
        """Returns circuit preparing the reference state of the ansatz (e.g
        prepare reference wavefunction with HF, multi-reference state, etc).
        These preparations must be consistent with the transform used to obtain
        the qubit operator.
        """

        return get_reference_circuit(n_spinorbitals=self.n_spinorbitals,
                                     n_electrons=self.n_electrons,
                                     mapping=self.qubit_mapping,
                                     up_then_down=self.up_then_down,
                                     spin=0)

    def build(self):
        """Builds the underlying objects required to run the QITE algorithm."""

        # Building molecule data with a pyscf molecule.
        if self.molecule:

            self.n_spinorbitals = self.molecule.n_active_sos
            self.n_electrons = self.molecule.n_active_electrons

            # Compute qubit hamiltonian for the input molecular system
            qubit_op = fermion_to_qubit_mapping(fermion_operator=self.molecule.fermionic_hamiltonian,
                                                mapping=self.qubit_mapping,
                                                n_spinorbitals=self.n_spinorbitals,
                                                n_electrons=self.n_electrons,
                                                up_then_down=self.up_then_down,
                                                spin=0)

            self.qubit_hamiltonian = qubitop_to_qubitham(qubit_op, self.qubit_mapping, self.up_then_down)

        # Getting the pool of operators for the ansatz. If more functionalities
        # are added, this part must be modified and generalized.
        if self.pool_args is None:
            if self.pool == uccgsd_pool:
                self.pool_args = (self.n_spinorbitals,)
            else:
                raise KeyError("pool_args must be defined if using own pool function")
        # Check if pool function returns a QubitOperator or FermionOperator and populate variables
        pool_list = self.pool(*self.pool_args)
        if isinstance(pool_list[0], QubitOperator):
            self.pool_type = "qubit"
            self.full_pool_operators = pool_list
        elif isinstance(pool_list[0], (FermionOperator, ofFermionOperator)):
            self.pool_type = "fermion"
            self.fermionic_operators = pool_list
            self.full_pool_operators = [fermion_to_qubit_mapping(fermion_operator=fi,
                                                                 mapping=self.qubit_mapping,
                                                                 n_spinorbitals=self.n_spinorbitals,
                                                                 n_electrons=self.n_electrons,
                                                                 up_then_down=self.up_then_down) for fi in self.fermionic_operators]
        else:
            raise ValueError("pool function must return either QubitOperator or FermionOperator")

        # Cast all coefs to floats (rotations angles are real).
        for qubit_op in self.full_pool_operators:
            for term, coeff in qubit_op.terms.items():
                qubit_op.terms[term] = math.copysign(1., coeff.imag)

        # Remove duplicates and only select terms with odd number of Y gates for all mappings except JKMN
        if self.qubit_mapping.upper() != "JKMN":
            reduced_pool_terms = set()
            for qubit_op in self.full_pool_operators:
                for term in qubit_op.terms:
                    count_y = str(term).count("Y")
                    if count_y % 2 == 1:
                        reduced_pool_terms.add(term)
        else:
            reduced_pool_terms = set()
            for qubit_op in self.full_pool_operators:
                for term in qubit_op.terms.keys():
                    if term:
                        reduced_pool_terms.add(term)

        # Generated list of pool_operators and full pool operator.
        self.pool_operators = [QubitOperator(term) for term in reduced_pool_terms]
        self.pool_qubit_op = QubitOperator()
        for term in self.pool_operators:
            self.pool_qubit_op += term

        self.qubit_operator = self.qubit_hamiltonian.to_qubitoperator()

        # Obtain all qubit terms that need to be measured
        self.pool_h = [element*self.qubit_operator for element in self.pool_operators]
        self.pool_pool = [[element1*element2 for element2 in self.pool_operators] for element1 in self.pool_operators]

        # Obtain initial state preparation circuit
        self.circuit_list.append(self.prepare_reference_state())
        self.final_circuit = copy(self.circuit_list[0])

        # Quantum circuit simulation backend options
        self.backend = Simulator(target=self.backend_options["target"], n_shots=self.backend_options["n_shots"],
                                 noise_model=self.backend_options["noise_model"])

        self.use_statevector = self.backend.statevector_available and self.backend._noise_model is None

    def simulate(self):
        """Performs the QITE cycles. Each iteration, a linear system is
        solved to obtain the next unitary. The system to be solved can be found in
        section 3.5 of https://arxiv.org/pdf/2108.04413.pdf

        Returns:
            float: final energy after obtaining running QITE
        """

        # Construction of the circuit. self.max_cycles terms are added, unless
        # the energy change is less than self.min_de.
        if self.use_statevector:
            self.update_statevector(self.backend, self.circuit_list[0])
        self.final_energy = self.energy_expectation(self.backend)
        self.energies.append(self.final_energy)
        while self.iteration < self.max_cycles:
            self.iteration += 1
            if self.verbose:
                print(f"Iteration {self.iteration} of QITE with starting energy {self.final_energy}")

            suv, bu = self.calculate_matrices(self.backend, self.final_energy)

            alphas_array = np.linalg.solve(suv.real, bu.real)
            # convert to dictionary with key as first (only) term of each pool_operator and value self.dt * alphas_array[i]
            alphas_dict = {next(iter(qu_op.terms)): self.dt * alphas_array[i] for i, qu_op in enumerate(self.pool_operators)}
            next_circuit = trotterize(self.pool_qubit_op, alphas_dict, trotter_order=1, n_trotter_steps=1)

            self.circuit_list.append(next_circuit)
            self.final_circuit += next_circuit

            if self.use_statevector:
                self.update_statevector(self.backend, self.circuit_list[self.iteration])

            new_energy = self.energy_expectation(self.backend)
            self.energies.append(new_energy)

            if abs(new_energy - self.final_energy) < self.min_de and self.iteration > 1:
                self.final_energy = new_energy
                break
            self.final_energy = new_energy

        if self.verbose:
            print(f"Final energy of QITE is {self.final_energy}")

        return self.energies[-1]

    def update_statevector(self, backend: Simulator, next_circuit: Circuit):
        r"""Update self.final_statevector by propagating with next_circuit using backend

        Args:
            Simulator: the backend to use for the statevector update
            Circuit: The circuit to apply to the statevector
        """
        _, self.final_statevector = backend.simulate(next_circuit,
                                                     return_statevector=True,
                                                     initial_statevector=self.final_statevector)

    def calculate_matrices(self, backend: Simulator, new_energy: float):
        r"""Calculated matrix elements for imaginary time evolution.
        The matrices are defined in section 3.5 of https://arxiv.org/pdf/2108.04413.pdf

        Args:
            backend (Simulator): the backend from which the matrices are generated
            new_energy (float): the current energy_expectation of the Hamiltonian

        Returns:
            matrix float: The expectation values <\psi| pu^+ pv |\psi>
            array float: The expecation values <\psi| pu^+ H |\psi>
        """

        circuit = Circuit(n_qubits=self.final_circuit.width) if self.use_statevector else self.final_circuit

        ndeltab = np.sqrt(1 - 2 * self.dt * new_energy)
        prefac = -1j/ndeltab
        bu = [prefac*backend.get_expectation_value(element, circuit, initial_statevector=self.final_statevector)
              for element in self.pool_h]
        bu = np.array(bu)
        pool_size = len(self.pool_h)
        suv = np.zeros((pool_size, pool_size), dtype=complex)
        for u in range(pool_size):
            for v in range(u+1, pool_size):
                suv[u, v] = backend.get_expectation_value(self.pool_pool[u][v],
                                                          circuit,
                                                          initial_statevector=self.final_statevector)
                suv[v, u] = suv[u, v]

        return suv, bu

    def energy_expectation(self, backend: Simulator):
        """Estimate energy using the self.final_circuit, qubit hamiltonian and compute
        backend.

        Args:
             backend (Simulator): the backend one computes the energy expectation with

        Returns:
            float: energy computed by the backend
        """
        circuit = Circuit(n_qubits=self.final_circuit.width) if self.use_statevector else self.final_circuit
        energy = backend.get_expectation_value(self.qubit_hamiltonian.to_qubitoperator(),
                                               circuit,
                                               initial_statevector=self.final_statevector)
        return energy

    def get_resources(self):
        """Returns resources currently used in underlying state preparation i.e. self.final_circuit
        the number of pool operators, and the size of qubit_hamiltonian

        Returns:
            dict: Dictionary of various quantum resources required"""
        resources = dict()
        resources["qubit_hamiltonian_terms"] = len(self.qubit_hamiltonian.terms)
        resources["pool_size"] = len(self.pool_operators)
        resources["circuit_width"] = self.final_circuit.width
        resources["circuit_gates"] = self.final_circuit.size
        resources["circuit_2qubit_gates"] = self.final_circuit.counts.get("CNOT", 0)
        return resources



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the construction of localized orbitals for DMET calculation."""

import unittest
import os
from pyscf import gto, scf
import numpy as np

from tangelo.problem_decomposition.dmet._helpers.dmet_orbitals import dmet_orbitals
from tangelo.problem_decomposition.electron_localization import iao_localization

path_file = os.path.dirname(os.path.abspath(__file__))


class TestDMETorbitals(unittest.TestCase):
    """Generate the localized orbitals employing IAOs."""

    def test_orbital_construction(self):

        # Initialize Molecule object with PySCF and input
        mol = gto.Mole()
        mol.atom = """
            C 0.94764 -0.02227  0.05901
            H 0.58322  0.35937 -0.89984
            H 0.54862  0.61702  0.85300
            H 0.54780 -1.03196  0.19694
            C 2.46782 -0.03097  0.07887
            H 2.83564  0.98716 -0.09384
            H 2.83464 -0.65291 -0.74596
            C 3.00694 -0.55965  1.40773
            H 2.63295 -1.57673  1.57731
            H 2.63329  0.06314  2.22967
            C 4.53625 -0.56666  1.42449
            H 4.91031  0.45032  1.25453
            H 4.90978 -1.19011  0.60302
            C 5.07544 -1.09527  2.75473
            H 4.70164 -2.11240  2.92450
            H 4.70170 -0.47206  3.57629
            C 6.60476 -1.10212  2.77147
            H 6.97868 -0.08532  2.60009
            H 6.97839 -1.72629  1.95057
            C 7.14410 -1.62861  4.10112
            H 6.77776 -2.64712  4.27473
            H 6.77598 -1.00636  4.92513
            C 8.66428 -1.63508  4.12154
            H 9.06449 -2.27473  3.32841
            H 9.02896 -2.01514  5.08095
            H 9.06273 -0.62500  3.98256
        """

        mol.basis = "3-21g"
        mol.charge = 0
        mol.spin = 0
        mol.build(verbose=0)

        mf = scf.RHF(mol)
        mf.scf()

        dmet_orbs = dmet_orbitals(mol, mf, range(mol.nao_nr()), iao_localization)
        dmet_orbitals_ref = np.loadtxt("{}/data/test_dmet_orbitals.txt".format(path_file))

        # Test the construction of IAOs
        for index, value_ref in np.ndenumerate(dmet_orbitals_ref):
            self.assertAlmostEqual(value_ref, dmet_orbs.localized_mo[index], msg="DMET orbs error at index " + str(index), delta=1e-6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools to define a reference state under a specified qubit-mapping, and
translate into a Circuit.
"""


import warnings

import numpy as np
from openfermion import QubitOperator
from openfermion.transforms import bravyi_kitaev_code
from openfermion.transforms.opconversions.bravyi_kitaev_tree import _transform_ladder_operator, FenwickTree

from tangelo.linq import Gate, Circuit
from tangelo.toolboxes.qubit_mappings.jkmn import jkmn_prep_vector

available_mappings = {"JW", "BK", "SCBK", "JKMN"}


def get_vector(n_spinorbitals, n_electrons, mapping, up_then_down=False, spin=None):
    """Get integer vector corresponding to Hartree Fock reference state.
    Reference state will occupy up to the n_electron-th molecular orbital.
    Depending on convention, basis is ordered alternating spin-up/spin-down
    (updown = False), or all up, then all down (updown = True).

    Args:
        n_spinorbitals (int): number of spin-orbitals in register.
        n_electrons (int): number of electrons in system.
        mapping (string): specify mapping, see mapping_transform.py for options
            "JW" (Jordan Wigner), or "BK" (Bravyi Kitaev), or "SCBK"
            (symmetry-conserving Bravyi Kitaev) or "JKMN"
            (Jiang Kalev Mruczkiewicz Neven)
        up_then_down (boolean): if True, all up, then all down, if False,
            alternating spin up/down.
         spin (int): 2*S = n_alpha - n_beta.

    Returns:
        numpy array of int: binary integer array indicating occupation of each
            spin-orbital.
    """
    if mapping.upper() not in available_mappings:
        raise ValueError(f"Invalid mapping selection. Select from: {available_mappings}")

    vector = np.zeros(n_spinorbitals, dtype=int)
    if spin:
        # if n_electrons is odd, then spin is also odd
        n_alpha = n_electrons//2 + spin//2 + (n_electrons % 2)
        n_beta = n_electrons//2 - spin//2
        vector[0:2*n_alpha:2] = 1
        vector[1:2*n_beta+1:2] = 1
    else:
        vector[:n_electrons] = 1
    return get_mapped_vector(vector, mapping, up_then_down)


def do_bk_transform(vector):
    """Apply Bravyi-Kitaev transformation to fermion occupation vector.
    Currently, simple wrapper on openfermion tools.

    Args:
        vector (numpy array of int): fermion occupation vector.

    Returns:
        numpy array of int: qubit-encoded occupation vector.
    """
    mat = bravyi_kitaev_code(len(vector)).encoder.toarray()
    vector_bk = np.mod(np.dot(mat, vector), 2)
    return vector_bk


def do_scbk_transform(vector, n_spinorbitals):
    """Instantiate qubit vector for symmetry-conserving Bravyi-Kitaev
    Generate Majorana mode for each occupied spin-orbital and apply X gate
    to each non-Z operator in the Pauli word.

    Args:
        vector (numpy array of int): fermion occupation vector.
        n_spinorbitals (int): number of qubits in register.

    Returns:
        numpy array of int: qubit-encoded occupation vector.
    """

    fenwick_tree = FenwickTree(n_spinorbitals)
    # Generate QubitOperator that represents excitation through Majorana mode (a_i^+ - a_) for each occupied orbital
    qu_op = QubitOperator((), 1)
    for ind, oc in enumerate(vector):
        if oc == 1:
            qu_op *= (_transform_ladder_operator((ind, 1), fenwick_tree) - _transform_ladder_operator((ind, 0), fenwick_tree))

    # Include all qubits that have Pauli operator X or Y acting on them in new occupation vector.
    vector_bk = np.zeros(n_spinorbitals)
    active_qus = [i for i, j in next(iter(qu_op.terms)) if j != "Z"]
    for q in active_qus:
        vector_bk[q] = 1

    # Delete n_spinorbital and last qubit as is done for the scBK transform.
    vector_bk = np.delete(vector_bk, n_spinorbitals-1)
    vector_scbk = np.delete(vector_bk, n_spinorbitals//2-1)
    return vector_scbk


def do_jkmn_transform(vector):
    """Instantiate qubit vector for JKMN transformation.

    Args:
        vector (numpy array of int): fermion occupation vector.

    Returns:
        numpy array of int: qubit-encoded occupation vector.
    """
    return jkmn_prep_vector(vector)


def get_mapped_vector(vector, mapping, up_then_down=False):
    """Return vector to generate circuit for a given occupation vector and mapping

    Args:
        vector (array of int): fermion occupation vector with up_then_down=False ordering.
            Number of spin-orbitals is assumed by length of array.
        mapping (str): One of the supported qubit mappings
        up_then_down (bool): if True, all up, then all down, if False,
            alternating spin up/down.

    Returns:
        array: The vector that generates the mapping occupations"""

    if up_then_down:
        vector = np.concatenate((vector[::2], vector[1::2]))
    if mapping.upper() == "JW":
        return vector
    elif mapping.upper() == "BK":
        return do_bk_transform(vector)
    elif mapping.upper() == "SCBK":
        if not up_then_down:
            warnings.warn("Symmetry-conserving Bravyi-Kitaev enforces all spin-up followed by all spin-down ordering.", RuntimeWarning)
            vector = np.concatenate((vector[::2], vector[1::2]))
        return do_scbk_transform(vector, len(vector))
    elif mapping.upper() == "JKMN":
        return do_jkmn_transform(vector)


def vector_to_circuit(vector):
    """Translate occupation vector into a circuit. Each occupied state
    corresponds to an X-gate on the associated qubit index.

    Args:
        vector (numpy array of int): occupation vector.

    Returns:
        Circuit: instance of tangelo.linq Circuit class.
    """

    n_qubits = len(vector)
    circuit = Circuit(n_qubits=n_qubits)

    for index, occupation in enumerate(vector):
        if occupation:
            gate = Gate("X", target=index)
            circuit.add_gate(gate)

    return circuit


def get_reference_circuit(n_spinorbitals, n_electrons, mapping, up_then_down=False, spin=None):
    """Build the Hartree-Fock state preparation circuit for the designated
    mapping.

    Args:
        n_spinorbitals (int): number of qubits in register.
        n_electrons (int): number of electrons in system.
        mapping (string): specify mapping, see mapping_transform.py for options
            "JW" (Jordan Wigner), or "BK" (Bravyi Kitaev), or "SCBK"
            (symmetry-conserving Bravyi Kitaev).
        up_then_down (boolean): if True, all up, then all down, if False,
            alternating spin up/down.
        spin (int): 2*S = n_alpha - n_beta.

    Returns:
        Circuit: instance of tangelo.linq Circuit class.
    """
    vector = get_vector(n_spinorbitals, n_electrons, mapping, up_then_down=up_then_down, spin=spin)
    circuit = vector_to_circuit(vector)
    return circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract class defining a common interface to all electronic structure
solvers, in order to have consistency and ensure compatibility with higher-level
workflows that can use them, such as problem decomposition methods.
"""

import abc


class ElectronicStructureSolver(abc.ABC):
    """Sets interface for objects that perform energy estimation of a molecule.

    Specifics vary between concrete implementations, but common to all of them
    is that after `simulate` is called, the right internal state is set for the
    class so that `get_rdm` can return reduced density matrices for the last run
    simulation.
    """

    def __init__(self, molecule):
        """Performs the simulation for a molecule. The mean field is calculated
        automatically.

        Args:
            molecule (SecondQuantizedMolecule): The molecule on which to perform the
                simulation.
        """
        pass

    @abc.abstractmethod
    def simulate(self):
        """Performs the simulation for a molecule."""
        pass

    @abc.abstractmethod
    def get_rdm(self):
        """Returns the RDM for the previous simulation.
        In a concrete implementation, the `simulate` function would set the
        necessary internal state for the class so that this function can return
        the reduced density matrix.

        Returns:
            numpy.array: One-particle RDM.
            numpy.array: Two-particle RDM.

        Raises:
            RuntimeError: If no simulation has been run.
        """
        pass



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Perform Meta-Löwdin localization.

The orbital localization of the canonical orbitals  using Meta-Löwdin
localization is done here. `pyscf.lo` is used.

For details, refer to:
    - Q. Sun et al., JCTC 10, 3784-3790 (2014).
"""

from pyscf.lo import orth


def meta_lowdin_localization(mol, mf):
    """Localize the orbitals using Meta-Löwdin localization.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate.
        mf (pyscf.scf.RHF): The mean field of the molecule.

    Returns:
        numpy.array: The localized orbitals (float64).
    """
    return orth.orth_ao(mol, "meta_lowdin")



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper function: pauli word rotations."""

import numpy as np
from tangelo.linq import Gate


def measurement_basis_gates(term):
    """Generate the rotation gates to perform change of basis before
    measurement.

    Args:
        term: Openfermion-style term. Essentially a list of (int, str) tuples.

    Returns:
        list of Gate: A list containing the rotation gates.
    """
    gates = []
    for qubit_index, pauli in term:
        if pauli in {"I", "Z"}:
            pass
        elif pauli == "X":
            gates.append(Gate("RY", qubit_index, parameter=-np.pi/2))
        elif pauli == "Y":
            gates.append(Gate("RX", qubit_index, parameter=np.pi/2))
        else:
            raise RuntimeError("Measurement basis not supported (currently supporting I,X,Y,Z)")
    return gates


def pauli_string_to_of(pauli_string):
    """ Converts a string of I,X,Y,Z Pauli operators to an Openfermion-style
    representation.
    """
    return [(i, p) for i, p in enumerate(pauli_string) if p != 'I']


def pauli_of_to_string(pauli_op, n_qubits):
    """ Converts an Openfermion-style Pauli word to a string representation.
    The user must specify the total number of qubits.
    """
    p_string = ['I'] * n_qubits
    for i, p in pauli_op:
        p_string[i] = p
    return ''.join(p_string)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This abstract quantum circuit class allows to represent a quantum circuit,
described by successive abstract quantum gate operations, without tying it to a
particular backend. It also provides methods to compute some of its
characteristics (width, size ...).
"""

import copy
from typing import List

import numpy as np

from tangelo.linq import Gate


class Circuit:
    """An abstract quantum circuit class, represented by a list of abstract gate
    operations acting on qubits indices. From this list of gates, the gate
    counts, width and other properties can be computed. In the future, this
    information allows for finding unentangled registers or qubits only left in
    a classical state.

    The attributes of the Circuit class should not be modified by the user: the
    add_gate method will ensure their value is correct as gates are added to the
    circuit object.

    It is however ok to modify the value of the parameters for variational gate
    operations in a sub-class of Circuit, as this does not impact the quantum
    circuit information. An example of how to do this to engineer a
    parameterized ansatz to be used in a variational algorithm in available in
    the example folder.
    """

    def __init__(self, gates: List[Gate] = None, n_qubits=None, name="no_name"):
        """Initialize gate list and internal variables depending on user input."""

        self.name = name
        self._gates = list()
        self._qubits_simulated = n_qubits
        self._qubit_indices = set() if not n_qubits else set(range(n_qubits))
        self._gate_counts = dict()
        self._variational_gates = []

        if gates:
            _ = [self.add_gate(g) for g in gates]

    def __str__(self):
        """Print info about the circuit and the gates it contains in a
        human-friendly format.
        """

        mystr = f"Circuit object. Size {self.size} \n\n"
        for abs_gate in self._gates:
            mystr += abs_gate.__str__() + "\n"
        return mystr

    def __add__(self, other):
        """Concatenate the list of instructions of two circuit objects into a
        single one.
        """
        return Circuit(self._gates + other._gates, n_qubits=max(self.width, other.width))

    def __mul__(self, n_repeat):
        """Return a circuit consisting of n_repeat repetitions of the input circuit.
        """
        if not isinstance(n_repeat, (int, np.integer)) or n_repeat <= 0:
            raise ValueError("Multiplication (repetition) operator with Circuit class only works for integers > 0")
        return Circuit(self._gates * n_repeat, n_qubits=self.width)

    def __rmul__(self, n_repeat):
        """Return a circuit consisting of n_repeat repetitions of the input circuit (circuit as right-hand side)
        """
        return self * n_repeat

    def __eq__(self, other):
        """Define equality (==) between 2 circuits. They are equal iff all their gates are equal, and they have
        the same numbers of qubits.
        """
        return (self._gates == other._gates) and (self.width == other.width)

    def __ne__(self, other):
        """Define inequality (!=) operator on circuits
        """
        return not (self == other)

    @property
    def size(self):
        """The size is the number of gates in the circuit. It is different from
        the depth.
        """
        return len(self._gates)

    @property
    def width(self):
        """Return the number of qubits required by this circuit. Assume all
        qubits are needed, even those who do not appear in a gate instruction.
        Qubit indexing starts at 0.
        """
        return max(self._qubit_indices) + 1 if self._qubit_indices else 0

    @property
    def counts(self):
        """Return the counts for all gate types included in the circuit."""
        return self._gate_counts

    @property
    def is_variational(self):
        """Returns true if the circuit holds any variational gate."""
        return True if self._variational_gates else False

    @property
    def is_mixed_state(self):
        """Assume circuit leads to a mixed state due to mid-circuit measurement
        if any MEASURE gate was explicitly added by the user.
        """
        return "MEASURE" in self.counts

    def add_gate(self, g):
        """Add a new gate to a circuit object and update other fields of the
        circuit object to gradually keep track of its properties (gate count,
        qubit indices...).
        """
        # Add a copy of the gate to the list of gates
        gate = Gate(g.name, g.target, g.control, g.parameter, g.is_variational)
        self._gates.append(gate)

        # A circuit is variational as soon as a variational gate is added to it
        if gate.is_variational:
            self._variational_gates.append(gate)

        def check_index_valid(index):
            """If circuit size was specified at instantiation, check that qubit
            indices do not go beyond it.
            """
            if self._qubits_simulated:
                if index >= self._qubits_simulated:
                    raise ValueError(f"Qubit index beyond expected maximal index ({self._qubits_simulated-1})\n "
                                     f"Gate = {gate}")

        # Track qubit indices
        all_involved_qubits = gate.target if gate.control is None else gate.target + gate.control
        for q in all_involved_qubits:
            check_index_valid(q)
            self._qubit_indices.add(q)

        # Keep track of the total gate count
        self._gate_counts[gate.name] = self._gate_counts.get(gate.name, 0) + 1

    def depth(self):
        """ Return the depth of the quantum circuit, by computing the number of moments. Does not count
        qubit initialization as a moment (unlike Cirq, for example).
        """
        moments = list()

        for g in self._gates:
            qubits = set(g.target) if g.control is None else set(g.target + g.control)

            if not moments:
                moments.append(qubits)
            else:
                # Find latest moment involving at least one of the qubits targeted by the current gate
                # The current gate is part of the moment following that one
                for i, m in reversed(list(enumerate(moments))):
                    if m & qubits:
                        if (i+1) < len(moments):
                            moments[i+1] = moments[i+1] | qubits
                        else:
                            moments.append(qubits)
                        break
                    # Case where none of the target qubits have been used before
                    elif i == 0:
                        moments[0] = moments[0] | qubits

        return len(moments)

    def trim_qubits(self):
        """Trim unnecessary qubits and update indices with the lowest values possible.
        """
        qubits_in_use = set().union(*self.get_entangled_indices())
        mapping = {ind: i for i, ind in enumerate(qubits_in_use)}
        for g in self._gates:
            g.target = [mapping[ind] for ind in g.target]
            if g.control:
                g.control = [mapping[ind] for ind in g.control]

        self._qubit_indices = set(range(len(qubits_in_use)))
        return self

    def reindex_qubits(self, new_indices):
        """Reindex qubit indices according to users labels / new indices. The
        new indices are set according to [new_index_for_qubit0,
        nex_index_for_qubit1, ..., new_index_for_qubit_N] where new_index < N.
        """

        if len(new_indices) != len(self._qubit_indices):
            raise ValueError(f"The number of indices does not match the length of self._qubit_indices")

        qubits_in_use = self._qubit_indices
        mapping = {i: j for i, j in zip(qubits_in_use, new_indices)}
        for g in self._gates:
            g.target = [mapping[ind] for ind in g.target]
            if g.control:
                g.control = [mapping[ind] for ind in g.control]

        self._qubit_indices = set(new_indices)

    def get_entangled_indices(self):
        """Return a list of unentangled sets of qubit indices. Each set includes indices
         of qubits that form an entangled subset.
        """

        entangled_indices = list()
        for g in self._gates:
            # Gradually accumulate entangled indices from the different subsets
            # Remove and replace them with their union, for each gate.
            q_new = set(g.target) if g.control is None else set(g.target + g.control)
            for qs in entangled_indices[::-1]:
                if q_new & qs:
                    q_new = q_new | qs
                    entangled_indices.remove(qs)
            entangled_indices.append(q_new)

        return entangled_indices

    def split(self):
        """ Split a circuit featuring unentangled qubit subsets into as many circuit objects.
        Each circuit only contains the gate operations targeting the qubit indices in its subsets.

        Returns:
            list of Circuit: list of resulting circuits
        """
        entangled_indices = self.get_entangled_indices()
        separate_circuits = [Circuit() for i in range(len(entangled_indices))]
        for g in self._gates:
            q_new = set(g.target) if g.control is None else set(g.target + g.control)
            # Append the gate to the circuit that handles the corresponding qubit indices
            for i, indices in enumerate(entangled_indices):
                if q_new & indices:
                    separate_circuits[i].add_gate(g)
                    break

        # Trim unnecessary indices in the new circuits
        for c in separate_circuits:
            c.trim_qubits()
        return separate_circuits

    def stack(self, *other_circuits):
        """Convenience method to stack other circuits on top of this one.
        See separate stack function.

        Args:
            *other_circuits (Circuit): one or several circuit objects to stack

        Returns:
            Circuit: the stacked circuit
        """
        return stack(self, *other_circuits)

    def inverse(self):
        """Return the inverse (adjoint) of a circuit

        This is performed by reversing the Gate order and applying the inverse to each Gate.

        Returns:
            Circuit: the inverted circuit
        """
        gates = [gate.inverse() for gate in reversed(self._gates)]
        return Circuit(gates, n_qubits=self.width)

    def serialize(self):
        if not isinstance(self.name, str):
            return TypeError("Name of circuit object must be a string")
        return {"name": self.name, "type": "QuantumCircuit", "gates": [gate.serialize() for gate in self._gates]}

    def remove_small_rotations(self, param_threshold=0.05):
        """Convenience method to remove small rotations from the circuit.
        See separate remove_small_rotations function.

        Args:
            param_threshold (float): Max absolute value to consider a rotation
                as a small one.

        Returns:
            Circuit: The circuit without small rotations.
        """
        opt_circuit = remove_small_rotations(self, param_threshold)
        self.__dict__ = opt_circuit.__dict__

    def remove_redundant_gates(self):
        """Convenience method to remove redundant gates from the circuit.
        See separate remove_redundant_gates function.

        Returns:
            Circuit: The circuit without redundant gates.
        """
        opt_circuit = remove_redundant_gates(self)
        self.__dict__ = opt_circuit.__dict__


def stack(*circuits):
    """ Take list of circuits as input, and stack them (e.g concatenate them along the
    width (qubits)) to form a single wider circuit, which allows users to run all of
    these circuits at once on a quantum device.

    Stacking provides a way to "fill up" a device if many qubits would be unused otherwise,
    therefore reducing cost / duration of a hardware experiment. However, depending on the
    device, this may amplify some sources of errors or make qubit placement more challenging.

    Args:
        *circuits (Circuit): the circuits to trim and stack into a single one

    Returns:
        Circuit: the stacked circuit
    """

    if not circuits:
        return Circuit()

    # Trim qubits of input circuit for maximum compactness
    circuits = [c.trim_qubits() for c in copy.deepcopy(list(circuits))]

    # Stack circuits. Reindex each circuit with the proper offset and then concatenate, until done
    stacked_circuit = circuits.pop(0)
    for c in circuits:
        c_stack = copy.deepcopy(c)
        c_stack.reindex_qubits(list(range(stacked_circuit.width, stacked_circuit.width + c.width)))
        stacked_circuit += c_stack

    return stacked_circuit


def remove_small_rotations(circuit, param_threshold=0.05):
    """Remove small rotation gates, up to a parameter threshold, from the
    circuit. Rotations from the set {"RX", "RY", "RZ", "CRX", "CRY", "CRZ"} are
    considered.

    Args:
        circuit (Circuit): the circuits to trim and stack into a single one
        param_threshold (float): Max absolute value to consider a rotation as
            a small one.

    Returns:
        Circuit: The circuit without small-rotation gates.
    """

    rot_gates = {"RX", "RY", "RZ", "CRX", "CRY", "CRZ"}
    gates = [g for g in circuit._gates if not (g.name in rot_gates and abs(g.parameter) % (2*np.pi) < param_threshold)]

    return Circuit(gates)


def remove_redundant_gates(circuit):
    """Remove redundant gates in a circuit. Redundant gates are adjacent gates
    that can be cancelled as their global effect is the identity. This function
    also works with many-qubit gates. However, it does not perform reordering of
    commutating gates to perform additional cancellations.

    Args:
        circuit (Circuit): the circuits to remove redundant gates.

    Returns:
        Circuit: The circuit without redundant gates.
    """
    gate_qubits = {i: list() for i in range(circuit.width)}
    indices_to_remove = list()

    for gi, gate in enumerate(circuit._gates):
        remove_gate = True

        # Identify qubits the current gate acts on.
        qubits = gate.target if gate.control is None else gate.target + gate.control

        # Check if the last gate cancels the current gate.
        for qubit_i in qubits:
            if not gate_qubits[qubit_i] or gate_qubits[qubit_i][-1][1].inverse() != gate:
                remove_gate = False
                break

        # Pop the last gate if the gate is to be removed.
        # If not, append the gate to the gate_qubits list.
        if remove_gate:
            indices_to_remove += [gi, gate_qubits[qubits[0]][-1][0]]
            for qubit_i in qubits:
                del gate_qubits[qubit_i][-1]
        else:
            for qubit_i in qubits:
                gate_qubits[qubit_i] += [(gi, gate)]

    # Remove gates that can be cancelled.
    gates = [gate for gate_i, gate in enumerate(circuit._gates) if gate_i not in indices_to_remove]

    return Circuit(gates)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for closed-shell and restricted open-shell qubit coupled cluster
with involutory linear combinations (ILC) of anticommuting sets (ACS) of Pauli words."""

import unittest

import numpy as np

from tangelo.linq import Simulator
from tangelo.toolboxes.ansatz_generator.ilc import ILC
from tangelo.toolboxes.ansatz_generator._qubit_ilc import gauss_elim_over_gf2
from tangelo.toolboxes.operators.operators import QubitOperator
from tangelo.molecule_library import mol_H2_sto3g, mol_H4_cation_sto3g

sim = Simulator()


class ILCTest(unittest.TestCase):
    """Unit tests for various functionalities of the ILC ansatz class. Examples for both closed-
    and restricted open-shell ILC are provided using H2, H4, and H4+.
    """

    @staticmethod
    def test_ilc_set_var_params():
        """ Verify behavior of set_var_params for different inputs (keyword, list, numpy array). """

        ilc_ansatz = ILC(mol_H2_sto3g, up_then_down=True)

        one_zero = np.zeros((1,), dtype=float)

        ilc_ansatz.set_var_params("qmf_state")
        np.testing.assert_array_almost_equal(ilc_ansatz.var_params, one_zero, decimal=6)

        ilc_ansatz.set_var_params([0.])
        np.testing.assert_array_almost_equal(ilc_ansatz.var_params, one_zero, decimal=6)

        one_tenth = 0.1 * np.ones((1,))

        ilc_ansatz.set_var_params([0.1])
        np.testing.assert_array_almost_equal(ilc_ansatz.var_params, one_tenth, decimal=6)

        ilc_ansatz.set_var_params(np.array([0.1]))
        np.testing.assert_array_almost_equal(ilc_ansatz.var_params, one_tenth, decimal=6)

    def test_ilc_incorrect_number_var_params(self):
        """ Return an error if user provide incorrect number of variational parameters """

        ilc_ansatz = ILC(mol_H2_sto3g, up_then_down=True)

        self.assertRaises(ValueError, ilc_ansatz.set_var_params, np.array([1.] * 2))

    @staticmethod
    def test_gauss_elim_over_gf2_sqrmat():
        """ Verify behavior of the Gaussian elimination over the binary field function. """

        # a_matrix stores the action of A * z over GF(2); dimension is n x m
        a_matrix = np.array([[1, 0, 1, 1], [1, 1, 0, 1], [1, 1, 1, 0], [0, 0, 1, 0]])

        # b_vec stores the solution vector for the equation A * z = b_vec; dimension is n x 1
        b_vec = np.array([1, 0, 1, 0]).reshape((4, 1))

        # z_ref stores the serves as the reference for the output of gauss_elim_over_gf2
        z_ref = np.array([0, 1, 0, 1])

        # solve A * z = b and compare to reference solution
        z_sln = gauss_elim_over_gf2(a_matrix, b_vec)

        np.testing.assert_array_almost_equal(z_sln, z_ref, decimal=6)

    @staticmethod
    def test_gauss_elim_over_gf2_rectmat():
        """ Verify behavior of the Gaussian elimination over the binary field function. """

        # a_matrix stores the action of A * z over GF(2); dimension is n x m
        a_matrix = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 0, 0], [0, 0, 0, 1, 1]])

        # b_vec stores the solution vector for the equation A * z = b_vec; dimension is n x 1
        b_vec = np.array([1, 1, 0]).reshape((3, 1))

        # z_ref stores the serves as the reference for the output of gauss_elim_over_gf2
        z_ref = np.array([1, 0, 1, 0, 0])

        # solve A * z = b and compare to reference solution
        z_sln = gauss_elim_over_gf2(a_matrix, b_vec)

        np.testing.assert_array_almost_equal(z_sln, z_ref, decimal=6)

    @staticmethod
    def test_gauss_elim_over_gf2_lindep():
        """ Verify behavior of the Gaussian elimination over the binary field function. """

        # a_matrix stores the action of A * z over GF(2); dimension is n x m
        a_matrix = np.array([[0, 0, 1, 0, 1], [0, 0, 1, 0, 1], [0, 0, 0, 1, 1]])

        # b_vec stores the solution vector for the equation A * z = b_vec; dimension is n x 1
        b_vec = np.array([1, 0, 1]).reshape((3, 1))

        # z_ref stores the serves as the reference for the output of gauss_elim_over_gf2
        z_ref = np.array([-1, -1, 1, 1, 0])

        # solve A * z = b and compare to reference solution
        z_sln = gauss_elim_over_gf2(a_matrix, b_vec)

        np.testing.assert_array_almost_equal(z_sln, z_ref, decimal=6)

    def test_ilc_h2(self):
        """ Verify closed-shell functionality when using the ILC class separately for H2 """

        # Specify the mutually anticommuting set (ACS) of ILC generators and parameters.
        acs = [QubitOperator("X0 Y1 Y2 Y3")]
        # The QMF parameters are automatically set if the argument qmf_var_params is not given.
        ilc_var_params = [0.11360304]
        ilc_ansatz = ILC(mol_H2_sto3g, "jw", True, acs)

        # Build the combined QMF (determined automatically if not specified) + ILC circuit.
        ilc_ansatz.build_circuit()

        # Get qubit hamiltonian for energy evaluation
        qubit_hamiltonian = ilc_ansatz.qubit_ham

        # Assert energy returned is as expected for given parameters
        ilc_ansatz.update_var_params(ilc_var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, ilc_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.1372697, delta=1e-6)

    def test_ilc_h4_cation(self):
        """ Verify restricted open-shell functionality when using the ILC class for H4+ """

        # Specify the mutually anticommuting set (ACS) of ILC generators and parameters.
        acs = [QubitOperator("Y0 Z2 X4 Z6"), QubitOperator("Y1 Y2 Z4 X5 Y6"), QubitOperator("X0 Z2 Z4 Y6"),
               QubitOperator("X1 Y2 X4 Z6"), QubitOperator("Y1 Y2 X4 Y5 Z6"), QubitOperator("Y1 Y2 Z4 Z5 Y6"),
               QubitOperator("Y0 Z1 Z2 Y5 Y6"), QubitOperator("Y0 Z1 Z2 Y4 Y5 Z6")]
        # The QMF parameters are automatically set if the argument qmf_var_params is not given.
        ilc_var_params = [ 0.14017492, -0.10792805, -0.05835484,  0.12468933,  0.07173118,  0.04683807,  0.02852163, -0.03133538]
        ilc_ansatz = ILC(mol_H4_cation_sto3g, "bk", False, acs)

        # Build the combined QMF (determined automatically if not specified) + ILC circuit.
        ilc_ansatz.build_circuit()

        # Get qubit hamiltonian for energy evaluation
        qubit_hamiltonian = ilc_ansatz.qubit_ham

        # Assert energy returned is as expected for given parameters
        ilc_ansatz.update_var_params(ilc_var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, ilc_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.6379638, delta=1e-6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utililty functions to generate pool of FermionOperators obtained from individual MajoranaOperators in
unitary coupled cluster expansions."""

from openfermion.transforms.opconversions.conversions import get_fermion_operator
from openfermion import MajoranaOperator
from numpy import integer

from tangelo.toolboxes.molecular_computation.molecule import SecondQuantizedMolecule


def get_majorana_uccsd_pool(molecule: SecondQuantizedMolecule = None, n_electrons: int = None, n_sos: int = None):
    """Construct a list of FermionOperator corresponding to the individual Majorana modes in the UCCSD ansatz

    Args:
        molecule (SecondQuantizedMolecule): The molecule to generate the Majorana pool from: Default None
        n_electrons (int): The number of active electrons: Default None
        n_sos (int): The number of active spin orbitals: Default None

    Returns:
        list: The list of FermionOperator for each Majorana operator in the UCCSD pool
    """

    if molecule is not None:
        n_active_electrons = molecule.n_active_electrons
        n_active_sos = molecule.n_active_sos
    elif isinstance(n_electrons, (int, integer)) and isinstance(n_sos, (int, integer)):
        n_active_electrons = n_electrons
        n_active_sos = n_sos
    else:
        raise ValueError("SecondQuantized mol or ints n_electrons/n_sos must be provided")

    def majorana_uccsd_generator():
        for i in range(n_active_electrons):
            for k in range(n_active_electrons, n_active_sos):
                yield (2*i, 2*k)
                yield (2*i+1, 2*k+1)
                for j in range(i+1, n_active_electrons):
                    for l in range(k+1, n_active_sos):
                        yield (2*i, 2*j+1, 2*k+1, 2*l+1)
                        yield (2*i+1, 2*j, 2*k+1, 2*l+1)
                        yield (2*i+1, 2*j+1, 2*k, 2*l+1)
                        yield (2*i+1, 2*j+1, 2*k+1, 2*l)
                        yield (2*i+1, 2*j, 2*k, 2*l)
                        yield (2*i, 2*j+1, 2*k, 2*l)
                        yield (2*i, 2*j, 2*k+1, 2*l)
                        yield (2*i, 2*j, 2*k, 2*l+1)

    pool_list = [get_fermion_operator(MajoranaOperator(term)) for term in majorana_uccsd_generator()]
    return pool_list


def get_majorana_uccgsd_pool(molecule: SecondQuantizedMolecule = None, n_sos: int = None):
    """Construct a list of FermionOperator corresponding to the individual Majorana modes in the UCCGSD ansatz

    Args:
        molecule (SecondQuantizedMolecule): The molecule to generate the Majorana pool from: Default None
        n_sos (int): The number of active spin orbitals: Default None

    Returns:
        list: The list of FermionOperator for each Majorana operator in the UCCGSD pool
    """

    if molecule is not None:
        n_active_sos = molecule.n_active_sos
    elif isinstance(n_sos, (int, integer)):
        n_active_sos = n_sos
    else:
        raise ValueError("SecondQuantized mol or int n_sos must be provided")

    def majorana_uccgsd_generator():
        for i in range(n_active_sos):
            for j in range(i+1, n_active_sos):
                yield (2*i, 2*j)
                yield (2*i+1, 2*j+1)
                for k in range(j+1, n_active_sos):
                    for l in range(k+1, n_active_sos):
                        yield (2*i, 2*j+1, 2*k+1, 2*l+1)
                        yield (2*i+1, 2*j, 2*k+1, 2*l+1)
                        yield (2*i+1, 2*j+1, 2*k, 2*l+1)
                        yield (2*i+1, 2*j+1, 2*k+1, 2*l)
                        yield (2*i+1, 2*j, 2*k, 2*l)
                        yield (2*i, 2*j+1, 2*k, 2*l)
                        yield (2*i, 2*j, 2*k+1, 2*l)
                        yield (2*i, 2*j, 2*k, 2*l+1)

    pool_list = [get_fermion_operator(MajoranaOperator(term)) for term in majorana_uccgsd_generator()]
    return pool_list
